/* //device/system/reference-ril/reference-ril.c
**
** Copyright 2006, The Android Open Source Project
**
** Licensed under the Apache License, Version 2.0 (the "License");
** you may not use this file except in compliance with the License.
** You may obtain a copy of the License at
**
**     http://www.apache.org/licenses/LICENSE-2.0
**
** Unless required by applicable law or agreed to in writing, software
** distributed under the License is distributed on an "AS IS" BASIS,
** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
** See the License for the specific language governing permissions and
** limitations under the License.
*/

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>
#include <alloca.h>
#include "atchannel.h"
#include "at_tok.h"
#include "misc.h"
#include <getopt.h>
#include <sys/socket.h>
#include <cutils/sockets.h>
#include <termios.h>
#include <sys/system_properties.h>
#if 1 //quectel
#include "../include/telephony/ril.h"
#include <cutils/properties.h>
#else

#include "ril.h"
#endif
#include "hardware/qemu_pipe.h"

#define LOG_TAG "RIL"
#include <utils/Log.h>

#if 1 //quectel PLATFORM_VERSION >= "4.2.2"
#ifndef LOGD
#define LOGD ALOGD
#endif
#ifndef LOGE
#define LOGE ALOGE
#endif
#ifndef LOGI
#define LOGI ALOGI
#endif
#endif

#define MAX_AT_RESPONSE 0x1000

/* pathname returned from RIL_REQUEST_SETUP_DATA_CALL / RIL_REQUEST_SETUP_DEFAULT_PDP */
#define PPP_TTY_PATH "eth0"

#ifdef USE_TI_COMMANDS

// Enable a workaround
// 1) Make incoming call, do not answer
// 2) Hangup remote end
// Expected: call should disappear from CLCC line
// Actual: Call shows as "ACTIVE" before disappearing
#define WORKAROUND_ERRONEOUS_ANSWER 1

// Some varients of the TI stack do not support the +CGEV unsolicited
// response. However, they seem to send an unsolicited +CME ERROR: 150
#define WORKAROUND_FAKE_CGEV 1
#endif

#if 1 //quectel
#undef WORKAROUND_FAKE_CGEV
#undef PPP_TTY_PATH
#define PPP_TTY_PATH "ppp0"
#define PPPD_EXIT_CODE           "net.gprs.ppp-exit"
#define PPP_START_RETRY_TIME  5
#define PPP_STOP_RETRY_TIME  5
#define CUSD_USE_UCS2_MODE
#define USB_HOST_SUSPEND_SUPPORT 0 //like s5pv210 donnot support usb suspend and wakeup
#define REFERENCE_RIL_VERSION    "Quectel_WCDMA&LTE_Android_RIL_SR01A23"
static  char ql_ttyAT[20];
extern int ql_get_ttyAT(char *out_ttyname);
static int gprsState = -1; // 1 ~in service, 0 ~ out of service
static int ql_is_UC20 = 0;
static int ql_is_UG95 = 0;
static int ql_is_EC20 = 0;
#define NETWORK_DEBOUNCE_TIMEOUT 20
static int network_debounce_time = 0;
static int onRequestCount = 0;
static int time_zone_report = 0;
#endif

typedef enum {
    SIM_ABSENT = 0,
    SIM_NOT_READY = 1,
    SIM_READY = 2, /* SIM_READY means the radio state is RADIO_STATE_SIM_READY */
    SIM_PIN = 3,
    SIM_PUK = 4,
    SIM_NETWORK_PERSONALIZATION = 5
} SIM_Status;

static void onRequest (int request, void *data, size_t datalen, RIL_Token t);
static RIL_RadioState currentState();
static int onSupports (int requestCode);
static void onCancel (RIL_Token t);
static const char *getVersion();
static int isRadioOn();
static SIM_Status getSIMStatus();
static int getCardStatus(RIL_CardStatus_v6 **pp_card_status);
static void freeCardStatus(RIL_CardStatus_v6 *p_card_status);
static void onDataCallListChanged(void *param);

extern const char * requestToString(int request);

/*** Static Variables ***/
static const RIL_RadioFunctions s_callbacks = {
    RIL_VERSION,
    onRequest,
    currentState,
    onSupports,
    onCancel,
    getVersion
};

#ifdef RIL_SHLIB
static const struct RIL_Env *s_rilenv;

#define RIL_onRequestComplete(t, e, response, responselen) s_rilenv->OnRequestComplete(t,e, response, responselen)
#define RIL_onUnsolicitedResponse(a,b,c) s_rilenv->OnUnsolicitedResponse(a,b,c)
#define RIL_requestTimedCallback(a,b,c) s_rilenv->RequestTimedCallback(a,b,c)
#endif

static RIL_RadioState sState = RADIO_STATE_UNAVAILABLE;

static pthread_mutex_t s_state_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t s_state_cond = PTHREAD_COND_INITIALIZER;

static int s_port = -1;
static const char * s_device_path = NULL;
static int          s_device_socket = 0;

/* trigger change to this with s_state_cond */
static int s_closed = 0;

static int sFD;     /* file desc of AT channel */
static char sATBuffer[MAX_AT_RESPONSE+1];
static char *sATBufferCur = NULL;

static const struct timeval TIMEVAL_SIMPOLL = {1,0};
static const struct timeval TIMEVAL_CALLSTATEPOLL = {0,500000};
static const struct timeval TIMEVAL_0 = {0,0};

#ifdef WORKAROUND_ERRONEOUS_ANSWER
// Max number of times we'll try to repoll when we think
// we have a AT+CLCC race condition
#define REPOLL_CALLS_COUNT_MAX 4

// Line index that was incoming or waiting at last poll, or -1 for none
static int s_incomingOrWaitingLine = -1;
// Number of times we've asked for a repoll of AT+CLCC
static int s_repollCallsCount = 0;
// Should we expect a call to be answered in the next CLCC?
static int s_expectAnswer = 0;
#endif /* WORKAROUND_ERRONEOUS_ANSWER */

static void pollSIMState (void *param);
static void setRadioState(RIL_RadioState newState);

#if 1 //quectel
static unsigned int ussd_pending_index = 0;
static const struct timeval ussd_timeout_timeval = {30,0}; //seconds
static void onUssdResponse(char *mode) {
    char *response[2]; 
    response[0] = mode;
       
    switch (mode[0] - '0') {
        case 2:
            response[1] = "USSD terminated by network";
        break;
        case 3:
            response[1] = "Other local client has responded";
        break;
        case 4:
            response[1] = "Operation not supported";
        break;
        case 5:
        default:
            response[1] = "Network time out";
        break;      
    }
    RIL_onUnsolicitedResponse(RIL_UNSOL_ON_USSD, response, sizeof(response[0]) + sizeof(response[1]));
}
static void onUssdTimedCallback(void *param) {
    if (ussd_pending_index != ((unsigned int) param))
        return;
    at_send_command("AT+CUSD=2", NULL); //cancel
    onUssdResponse("5");
}

static int gsm_hexchar_to_int( char  c ) {
    if ((unsigned)(c - '0') < 10)
        return c - '0';
    if ((unsigned)(c - 'a') < 6)
        return 10 + (c - 'a');
    if ((unsigned)(c - 'A') < 6)
        return 10 + (c - 'A');
    return -1;
}

static int gsm_hex2_to_byte( const char*  hex ) {
    int  hi = gsm_hexchar_to_int(hex[0]);
    int  lo = gsm_hexchar_to_int(hex[1]);

    if (hi < 0 || lo < 0)
        return -1;

    return ( (hi << 4) | lo );
}

typedef unsigned char  byte_t;
typedef byte_t*        bytes_t;
typedef const byte_t*  cbytes_t;
static  int utf8_write( bytes_t  utf8, int  offset, int  v ) {
    int  result;

    if (v < 128) {
        result = 1;
        if (utf8)
            utf8[offset] = (byte_t) v;
    } else if (v < 0x800) {
        result = 2;
        if (utf8) {
            utf8[offset+0] = (byte_t)( 0xc0 | (v >> 6) );
            utf8[offset+1] = (byte_t)( 0x80 | (v & 0x3f) );
        }
    } else if (v < 0x10000) {
        result = 3;
        if (utf8) {
            utf8[offset+0] = (byte_t)( 0xe0 |  (v >> 12) );
            utf8[offset+1] = (byte_t)( 0x80 | ((v >> 6) & 0x3f) );
            utf8[offset+2] = (byte_t)( 0x80 |  (v & 0x3f) );
        }
    } else {
        result = 4;
        if (utf8) {
            utf8[offset+0] = (byte_t)( 0xf0 | ((v >> 18) & 0x7) );
            utf8[offset+1] = (byte_t)( 0x80 | ((v >> 12) & 0x3f) );
            utf8[offset+2] = (byte_t)( 0x80 | ((v >> 6) & 0x3f) );
            utf8[offset+3] = (byte_t)( 0x80 |  (v & 0x3f) );
        }
    }
    return  result;
}

/* convert a UCS2 string into a UTF8 byte string, assumes 'buf' is correctly sized */
static int ucs2_to_utf8( cbytes_t  ucs2, int       ucs2len, bytes_t   buf ) {
    int  nn;
    int  result = 0;

    for (nn = 0; nn < ucs2len; ucs2 += 2, nn++) {
        int  c= (ucs2[0] << 8) | ucs2[1];
        result += utf8_write(buf, result, c);
    }
    return result;
}
#endif

static int clccStateToRILState(int state, RIL_CallState *p_state)

{
    switch(state) {
        case 0: *p_state = RIL_CALL_ACTIVE;   return 0;
        case 1: *p_state = RIL_CALL_HOLDING;  return 0;
        case 2: *p_state = RIL_CALL_DIALING;  return 0;
        case 3: *p_state = RIL_CALL_ALERTING; return 0;
        case 4: *p_state = RIL_CALL_INCOMING; return 0;
        case 5: *p_state = RIL_CALL_WAITING;  return 0;
        default: return -1;
    }
}

/**
 * Note: directly modified line and has *p_call point directly into
 * modified line
 */
static int callFromCLCCLine(char *line, RIL_Call *p_call)
{
        //+CLCC: 1,0,2,0,0,\"+18005551212\",145
        //     index,isMT,state,mode,isMpty(,number,TOA)?

    int err;
    int state;
    int mode;

    err = at_tok_start(&line);
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &(p_call->index));
    if (err < 0) goto error;

    err = at_tok_nextbool(&line, &(p_call->isMT));
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &state);
    if (err < 0) goto error;

    err = clccStateToRILState(state, &(p_call->state));
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &mode);
    if (err < 0) goto error;

    p_call->isVoice = (mode == 0);

    err = at_tok_nextbool(&line, &(p_call->isMpty));
    if (err < 0) goto error;

    if (at_tok_hasmore(&line)) {
        err = at_tok_nextstr(&line, &(p_call->number));

        /* tolerate null here */
        if (err < 0) return 0;

        // Some lame implementations return strings
        // like "NOT AVAILABLE" in the CLCC line
        if (p_call->number != NULL
            && 0 == strspn(p_call->number, "+0123456789")
        ) {
            p_call->number = NULL;
        }

        err = at_tok_nextint(&line, &p_call->toa);
        if (err < 0) goto error;
    }

    p_call->uusInfo = NULL;

    return 0;

error:
    LOGE("invalid CLCC line\n");
    return -1;
}


/** do post-AT+CFUN=1 initialization */
static void onRadioPowerOn()
{
#ifdef USE_TI_COMMANDS
    /*  Must be after CFUN=1 */
    /*  TI specific -- notifications for CPHS things such */
    /*  as CPHS message waiting indicator */

    at_send_command("AT%CPHS=1", NULL);

    /*  TI specific -- enable NITZ unsol notifs */
    at_send_command("AT%CTZV=1", NULL);
#endif

    pollSIMState(NULL);
}

/** do post- SIM ready initialization */
static void onSIMReady()
{
    at_send_command_singleline("AT+CSMS=1", "+CSMS:", NULL);
    /*
     * Always send SMS messages directly to the TE
     *
     * mode = 1 // discard when link is reserved (link should never be
     *             reserved)
     * mt = 2   // most messages routed to TE
     * bm = 2   // new cell BM's routed to TE
     * ds = 1   // Status reports routed to TE
     * bfr = 1  // flush buffer
     */
#if 1 //quectel
    at_send_command("AT+CNMI=1,2,0,1,0", NULL);
#else
    at_send_command("AT+CNMI=1,2,2,1,1", NULL);
#endif
}

static void requestRadioPower(void *data, size_t datalen, RIL_Token t)
{
    int onOff;

    int err;
    ATResponse *p_response = NULL;
    
    assert (datalen >= sizeof(int *));
    onOff = ((int *)data)[0];

    if (onOff == 0 && sState != RADIO_STATE_OFF) {
 #if 1 //quectel        
       /**
         *  Wythe modify on 2013-10-8
         *  just switch off the RF. CFUN=0 will shutdown the simcard.
         */
        err = at_send_command("AT+CFUN=4", &p_response);
#else   
        err = at_send_command("AT+CFUN=0", &p_response);
#endif
       if (err < 0 || p_response->success == 0) goto error;
        setRadioState(RADIO_STATE_OFF);
    } else if (onOff > 0 && sState == RADIO_STATE_OFF) {
        err = at_send_command("AT+CFUN=1", &p_response);
        if (err < 0|| p_response->success == 0) {     
            // Some stacks return an error when there is no SIM,
            // but they really turn the RF portion on
            // So, if we get an error, let's check to see if it
            // turned on anyway
            
            if (isRadioOn() != 1) {
                goto error;
            }
        }
        setRadioState(RADIO_STATE_SIM_NOT_READY);
    }

    at_response_free(p_response);
    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    return;
error:
    LOGE("[%s] error", __func__);
    at_response_free(p_response);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void requestOrSendDataCallList(RIL_Token *t);

static void onDataCallListChanged(void *param)
{
    requestOrSendDataCallList(NULL);
}

static void requestDataCallList(void *data, size_t datalen, RIL_Token t)
{
    requestOrSendDataCallList(&t);
}

static void requestOrSendDataCallList(RIL_Token *t)
{
    ATResponse *p_response;
    ATLine *p_cur;
    int err;
    int n = 0;
    char *out;

    err = at_send_command_multiline ("AT+CGACT?", "+CGACT:", &p_response);
    if (err != 0 || p_response->success == 0) {
        if (t != NULL)
            RIL_onRequestComplete(*t, RIL_E_GENERIC_FAILURE, NULL, 0);
        else
            RIL_onUnsolicitedResponse(RIL_UNSOL_DATA_CALL_LIST_CHANGED,
                                      NULL, 0);
        return;
    }

    for (p_cur = p_response->p_intermediates; p_cur != NULL;
         p_cur = p_cur->p_next)
        n++;

    RIL_Data_Call_Response_v6 *responses =
        alloca(n * sizeof(RIL_Data_Call_Response_v6));

    int i;
    for (i = 0; i < n; i++) {
        responses[i].status = -1;
        responses[i].suggestedRetryTime = -1;
        responses[i].cid = -1;
        responses[i].active = -1;
#if 1 //quectel
        responses[i].type = (char *)"";
        responses[i].ifname = (char *)PPP_TTY_PATH;
        responses[i].addresses = (char *)"";
        responses[i].dnses = (char *)"";
        responses[i].gateways = (char *)"";
#else
        responses[i].type = "";
        responses[i].ifname = "";
        responses[i].addresses = "";
        responses[i].dnses = "";
        responses[i].gateways = "";
#endif
    }

    RIL_Data_Call_Response_v6 *response = responses;
    for (p_cur = p_response->p_intermediates; p_cur != NULL;
         p_cur = p_cur->p_next) {
        char *line = p_cur->line;

        err = at_tok_start(&line);
        if (err < 0)
            goto error;

        err = at_tok_nextint(&line, &response->cid);
        if (err < 0)
            goto error;

        err = at_tok_nextint(&line, &response->active);
        if (err < 0)
            goto error;

        response++;
    }

    at_response_free(p_response);

    err = at_send_command_multiline ("AT+CGDCONT?", "+CGDCONT:", &p_response);
    if (err != 0 || p_response->success == 0) {
        if (t != NULL)
            RIL_onRequestComplete(*t, RIL_E_GENERIC_FAILURE, NULL, 0);
        else
            RIL_onUnsolicitedResponse(RIL_UNSOL_DATA_CALL_LIST_CHANGED,
                                      NULL, 0);
        return;
    }

    for (p_cur = p_response->p_intermediates; p_cur != NULL;
         p_cur = p_cur->p_next) {
        char *line = p_cur->line;
        int cid;

        err = at_tok_start(&line);
        if (err < 0)
            goto error;

        err = at_tok_nextint(&line, &cid);
        if (err < 0)
            goto error;

        for (i = 0; i < n; i++) {
            if (responses[i].cid == cid)
                break;
        }

        if (i >= n) {
            /* details for a context we didn't hear about in the last request */
            continue;
        }

        // Assume no error
        responses[i].status = 0;

        // type
        err = at_tok_nextstr(&line, &out);
        if (err < 0)
            goto error;
        responses[i].type = alloca(strlen(out) + 1);
        strcpy(responses[i].type, out);

        // APN ignored for v5
        err = at_tok_nextstr(&line, &out);
        if (err < 0)
            goto error;

        responses[i].ifname = alloca(strlen(PPP_TTY_PATH) + 1);
        strcpy(responses[i].ifname, PPP_TTY_PATH);

        err = at_tok_nextstr(&line, &out);
        if (err < 0)
            goto error;

        responses[i].addresses = alloca(strlen(out) + 1);
        strcpy(responses[i].addresses, out);

        {
            char  propValue[PROP_VALUE_MAX];

            if (__system_property_get("ro.kernel.qemu", propValue) != 0) {
                /* We are in the emulator - the dns servers are listed
                 * by the following system properties, setup in
                 * /system/etc/init.goldfish.sh:
                 *  - net.eth0.dns1
                 *  - net.eth0.dns2
                 *  - net.eth0.dns3
                 *  - net.eth0.dns4
                 */
                const int   dnslist_sz = 128;
                char*       dnslist = alloca(dnslist_sz);
                const char* separator = "";
                int         nn;

                dnslist[0] = 0;
                for (nn = 1; nn <= 4; nn++) {
                    /* Probe net.eth0.dns<n> */
                    char  propName[PROP_NAME_MAX];
                    snprintf(propName, sizeof propName, "net.eth0.dns%d", nn);

                    /* Ignore if undefined */
                    if (__system_property_get(propName, propValue) == 0) {
                        continue;
                    }

                    /* Append the DNS IP address */
                    strlcat(dnslist, separator, dnslist_sz);
                    strlcat(dnslist, propValue, dnslist_sz);
                    separator = " ";
                }
                responses[i].dnses = dnslist;

                /* There is only on gateway in the emulator */
                responses[i].gateways = "10.0.2.2";
            }
            else {
                /* I don't know where we are, so use the public Google DNS
                 * servers by default and no gateway.
                 */
                responses[i].dnses = "8.8.8.8 8.8.4.4";
                responses[i].gateways = "";
            }
        }
    }

    at_response_free(p_response);

    if (t != NULL)
        RIL_onRequestComplete(*t, RIL_E_SUCCESS, responses,
                              n * sizeof(RIL_Data_Call_Response_v6));
    else
        RIL_onUnsolicitedResponse(RIL_UNSOL_DATA_CALL_LIST_CHANGED,
                                  responses,
                                  n * sizeof(RIL_Data_Call_Response_v6));

    return;

error:
    LOGE("%s error", __func__);
    if (t != NULL)
        RIL_onRequestComplete(*t, RIL_E_GENERIC_FAILURE, NULL, 0);
    else
        RIL_onUnsolicitedResponse(RIL_UNSOL_DATA_CALL_LIST_CHANGED,
                                  NULL, 0);

    at_response_free(p_response);
}

static void requestQueryNetworkSelectionMode(
                void *data, size_t datalen, RIL_Token t)
{
    int err;
    ATResponse *p_response = NULL;
    int response = 0;
    char *line;

    err = at_send_command_singleline("AT+COPS?", "+COPS:", &p_response);

    if (err < 0 || p_response->success == 0) {
        goto error;
    }

    line = p_response->p_intermediates->line;

    err = at_tok_start(&line);

    if (err < 0) {
        goto error;
    }

    err = at_tok_nextint(&line, &response);

    if (err < 0) {
        goto error;
    }

    RIL_onRequestComplete(t, RIL_E_SUCCESS, &response, sizeof(int));
    at_response_free(p_response);
    return;
error:
    LOGE("%s error", __func__);
    at_response_free(p_response);
    LOGE("requestQueryNetworkSelectionMode must never return error when radio is on");
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void sendCallStateChanged(void *param)
{
    RIL_onUnsolicitedResponse (
        RIL_UNSOL_RESPONSE_CALL_STATE_CHANGED,
        NULL, 0);
}

static void requestGetCurrentCalls(void *data, size_t datalen, RIL_Token t)
{
    int err;
    ATResponse *p_response;
    ATLine *p_cur;
    int countCalls;
    int countValidCalls;
    RIL_Call *p_calls;
    RIL_Call **pp_calls;
    int i;
    int needRepoll = 0;

#ifdef WORKAROUND_ERRONEOUS_ANSWER
    int prevIncomingOrWaitingLine;

    prevIncomingOrWaitingLine = s_incomingOrWaitingLine;
    s_incomingOrWaitingLine = -1;
#endif /*WORKAROUND_ERRONEOUS_ANSWER*/

    err = at_send_command_multiline ("AT+CLCC", "+CLCC:", &p_response);

    if (err != 0 || p_response->success == 0) {
        LOGE("%s response generic failure", __func__);
        RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
        return;
    }

    /* count the calls */
    for (countCalls = 0, p_cur = p_response->p_intermediates
            ; p_cur != NULL
            ; p_cur = p_cur->p_next
    ) {
        countCalls++;
    }

    /* yes, there's an array of pointers and then an array of structures */

    pp_calls = (RIL_Call **)alloca(countCalls * sizeof(RIL_Call *));
    p_calls = (RIL_Call *)alloca(countCalls * sizeof(RIL_Call));
    memset (p_calls, 0, countCalls * sizeof(RIL_Call));

    /* init the pointer array */
    for(i = 0; i < countCalls ; i++) {
        pp_calls[i] = &(p_calls[i]);
    }

    for (countValidCalls = 0, p_cur = p_response->p_intermediates
            ; p_cur != NULL
            ; p_cur = p_cur->p_next
    ) {
        err = callFromCLCCLine(p_cur->line, p_calls + countValidCalls);

        if (err != 0) {
            continue;
        }

#ifdef WORKAROUND_ERRONEOUS_ANSWER
        if (p_calls[countValidCalls].state == RIL_CALL_INCOMING
            || p_calls[countValidCalls].state == RIL_CALL_WAITING
        ) {
            s_incomingOrWaitingLine = p_calls[countValidCalls].index;
        }
#endif /*WORKAROUND_ERRONEOUS_ANSWER*/

        if (p_calls[countValidCalls].state != RIL_CALL_ACTIVE
            && p_calls[countValidCalls].state != RIL_CALL_HOLDING
        ) {
            needRepoll = 1;
        }
#if 1 //quectel
        if(p_calls[countValidCalls].isVoice)
#endif
        countValidCalls++;
    }

#ifdef WORKAROUND_ERRONEOUS_ANSWER
    // Basically:
    // A call was incoming or waiting
    // Now it's marked as active
    // But we never answered it
    //
    // This is probably a bug, and the call will probably
    // disappear from the call list in the next poll
    if (prevIncomingOrWaitingLine >= 0
            && s_incomingOrWaitingLine < 0
            && s_expectAnswer == 0
    ) {
        for (i = 0; i < countValidCalls ; i++) {

            if (p_calls[i].index == prevIncomingOrWaitingLine
                    && p_calls[i].state == RIL_CALL_ACTIVE
                    && s_repollCallsCount < REPOLL_CALLS_COUNT_MAX
            ) {
                LOGI(
                    "Hit WORKAROUND_ERRONOUS_ANSWER case."
                    " Repoll count: %d\n", s_repollCallsCount);
                s_repollCallsCount++;
                goto error;
            }
        }
    }

    s_expectAnswer = 0;
    s_repollCallsCount = 0;
#endif /*WORKAROUND_ERRONEOUS_ANSWER*/

    LOGI("Calls=%d,Valid=%d",countCalls,countValidCalls);
    
    RIL_onRequestComplete(t, RIL_E_SUCCESS, pp_calls,
            countValidCalls * sizeof (RIL_Call *));

    at_response_free(p_response);

#ifdef POLL_CALL_STATE
    if (countValidCalls) {  // We don't seem to get a "NO CARRIER" message from
                            // smd, so we're forced to poll until the call ends.
#else
    if (needRepoll) {
#endif
        RIL_requestTimedCallback (sendCallStateChanged, NULL, &TIMEVAL_CALLSTATEPOLL);
    }

    return;
error:
    LOGD("%s error", __func__);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}

static void requestDial(void *data, size_t datalen, RIL_Token t)
{
    RIL_Dial *p_dial;
    char *cmd;
    const char *clir;
    int ret;

    p_dial = (RIL_Dial *)data;

    switch (p_dial->clir) {
        case 1: clir = "I"; break;  /*invocation*/
        case 2: clir = "i"; break;  /*suppression*/
        default:
        case 0: clir = ""; break;   /*subscription default*/
    }

    asprintf(&cmd, "ATD%s%s;", p_dial->address, clir);

    ret = at_send_command(cmd, NULL);

    free(cmd);

    /* success or failure is ignored by the upper layer here.
       it will call GET_CURRENT_CALLS and determine success that way */
    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}

static void requestWriteSmsToSim(void *data, size_t datalen, RIL_Token t)
{
    RIL_SMS_WriteArgs *p_args;
    char *cmd;
    int length;
    int err;
    ATResponse *p_response = NULL;

    p_args = (RIL_SMS_WriteArgs *)data;

    length = strlen(p_args->pdu)/2;
    asprintf(&cmd, "AT+CMGW=%d,%d", length, p_args->status);

    err = at_send_command_sms(cmd, p_args->pdu, "+CMGW:", &p_response);

    if (err != 0 || p_response->success == 0) goto error;

    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    at_response_free(p_response);

    return;
error:
    LOGE("%s error\n", __func__);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}

static void requestHangup(void *data, size_t datalen, RIL_Token t)
{
    int *p_line;

    int ret;
    char *cmd;

    p_line = (int *)data;

    // 3GPP 22.030 6.5.5
    // "Releases a specific active call X"
    asprintf(&cmd, "AT+CHLD=1%d", p_line[0]);

    ret = at_send_command(cmd, NULL);

    free(cmd);

    /* success or failure is ignored by the upper layer here.
       it will call GET_CURRENT_CALLS and determine success that way */
    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}

static void requestSignalStrength(void *data, size_t datalen, RIL_Token t)
{
    ATResponse *p_response = NULL;
    int err;
    int response[2];
    char *line;

    err = at_send_command_singleline("AT+CSQ", "+CSQ:", &p_response);

    if (err < 0 || p_response->success == 0) {
        RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
        goto error;
    }

    line = p_response->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &(response[0]));
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &(response[1]));
    if (err < 0) goto error;

#if 1 //quectel
{
    RIL_SignalStrength_v6 signalStrength;
    memset(&signalStrength, 0, sizeof(RIL_SignalStrength_v6));
    /**
     *  In Android Jelly Bean, the invalid value for
     *  LET signalStrength should be 99 depending on the SignalStrength.java.
     */
//#if ((PLATFORM_VERSION >= 420) || ((PLATFORM_VERSION < 100) && (PLATFORM_VERSION >= 42)))
#if (RIL_VERSION >= 7)
    signalStrength.LTE_SignalStrength.signalStrength = 99;
    signalStrength.LTE_SignalStrength.rsrp = 0x7FFFFFFF;
    signalStrength.LTE_SignalStrength.rsrq = 0x7FFFFFFF;
    signalStrength.LTE_SignalStrength.rssnr = 0x7FFFFFFF;
    signalStrength.LTE_SignalStrength.cqi = 0x7FFFFFFF;
#else
    signalStrength.LTE_SignalStrength.signalStrength = -1;
    signalStrength.LTE_SignalStrength.rsrp = -1;
    signalStrength.LTE_SignalStrength.rsrq = -1;
    signalStrength.LTE_SignalStrength.rssnr = -1;
    signalStrength.LTE_SignalStrength.cqi = -1;
#endif	
    signalStrength.GW_SignalStrength.signalStrength = response[0];
    signalStrength.GW_SignalStrength.bitErrorRate = response[1];
    RIL_onRequestComplete(t, RIL_E_SUCCESS, &signalStrength, sizeof(RIL_SignalStrength_v6));
}
#else
    RIL_onRequestComplete(t, RIL_E_SUCCESS, response, sizeof(response));
#endif
    at_response_free(p_response);
    return;

error:
    LOGE("requestSignalStrength must never return an error when radio is on");
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}

#if 1 //quectel
static int quectel_get_Act_from_cops_for_cgreg(void){
    int err;
    ATResponse *p_response = NULL;
    int response = 0;
    char *line;
    int mode, format, act = 0;
    char *oper;

    err = at_send_command_singleline("AT+COPS?", "+COPS:", &p_response);

    if (err < 0 || p_response->success == 0) {
        goto error;
    }

    line = p_response->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0) goto error;

//+COPS:<mode>[,<format>[,<oper>][,<Act>]]
    err = at_tok_nextint(&line, &mode);
    if (err < 0) goto error;

    if (!at_tok_hasmore(&line)) goto error;

    err = at_tok_nextint(&line, &format);
    if (err < 0) goto error;

    if (!at_tok_hasmore(&line)) goto error;

    err = at_tok_nextstr(&line, &oper);
    if (err < 0) goto error;

    if (!at_tok_hasmore(&line)) goto error;

    err = at_tok_nextint(&line, &act);
    if (err < 0) goto error;

error:
     at_response_free(p_response);
     return act;
}
#endif

static void requestRegistrationState(int request, void *data,
                                        size_t datalen, RIL_Token t)
{
    int err;
    int response[4];
    char * responseStr[4];
    ATResponse *p_response = NULL;
    const char *cmd;
    const char *prefix;
    char *line, *p;
    int commas;
    int skip;
    int count = 3;


    if (request == RIL_REQUEST_VOICE_REGISTRATION_STATE) {
        cmd = "AT+CREG?";
        prefix = "+CREG:";
    } else if (request == RIL_REQUEST_DATA_REGISTRATION_STATE) {
        cmd = "AT+CGREG?";
        prefix = "+CGREG:";
    } else {
        assert(0);
        goto error;
    }

#if 1 //quectel
__requestRegistrationState_restart:
    skip = 2; //unsolicited result code with location information +CGREG: 2,<stat>[,<lac>,<ci>[,<Act>]]
#endif
    err = at_send_command_singleline(cmd, prefix, &p_response);

    if (err != 0) goto error;

    line = p_response->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0) goto error;

    /* Ok you have to be careful here
     * The solicited version of the CREG response is
     * +CREG: n, stat, [lac, cid]
     * and the unsolicited version is
     * +CREG: stat, [lac, cid]
     * The <n> parameter is basically "is unsolicited creg on?"
     * which it should always be
     *
     * Now we should normally get the solicited version here,
     * but the unsolicited version could have snuck in
     * so we have to handle both
     *
     * Also since the LAC and CID are only reported when registered,
     * we can have 1, 2, 3, or 4 arguments here
     *
     * finally, a +CGREG: answer may have a fifth value that corresponds
     * to the network type, as in;
     *
     *   +CGREG: n, stat [,lac, cid [,networkType]]
     */

    /* count number of commas */
    commas = 0;
    for (p = line ; *p != '\0' ;p++) {
        if (*p == ',') commas++;
    }

    switch (commas) {
        case 0: /* +CREG: <stat> */
            err = at_tok_nextint(&line, &response[0]);
            if (err < 0) goto error;
            response[1] = -1;
            response[2] = -1;
        break;

        case 1: /* +CREG: <n>, <stat> */
            err = at_tok_nextint(&line, &skip);
            if (err < 0) goto error;
            err = at_tok_nextint(&line, &response[0]);
            if (err < 0) goto error;
            response[1] = -1;
            response[2] = -1;
            if (err < 0) goto error;
        break;

        case 2: /* +CREG: <stat>, <lac>, <cid> */
            err = at_tok_nextint(&line, &response[0]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &response[1]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &response[2]);
            if (err < 0) goto error;
        break;
        case 3: /* +CREG: <n>, <stat>, <lac>, <cid> */
            err = at_tok_nextint(&line, &skip);
            if (err < 0) goto error;
            err = at_tok_nextint(&line, &response[0]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &response[1]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &response[2]);
            if (err < 0) goto error;
        break;
        /* special case for CGREG, there is a fourth parameter
         * that is the network type (unknown/gprs/edge/umts)
         */
        case 4: /* +CGREG: <n>, <stat>, <lac>, <cid>, <networkType> */
            err = at_tok_nextint(&line, &skip);
            if (err < 0) goto error;
            err = at_tok_nextint(&line, &response[0]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &response[1]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &response[2]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &response[3]);
            if (err < 0) goto error;
#if 1 //quectel UC15&UC20&UG95 canot get raido tech from AT+CGREG?, what about EC20?
            if ((request == RIL_REQUEST_DATA_REGISTRATION_STATE) && (response[3] == 2)) {
                int real_act = quectel_get_Act_from_cops_for_cgreg();
                if (real_act > response[3])
                    response[3] = real_act;
            }
#endif
            count = 4;
        break;
        default:
            goto error;
    }

    asprintf(&responseStr[0], "%d", response[0]);
    asprintf(&responseStr[1], "%x", response[1]);
    asprintf(&responseStr[2], "%x", response[2]);

#if 1 //quectel
    if (skip != 2) {
        at_response_free(p_response);
        p_response = NULL;
        at_send_command("AT+CREG=2", NULL); 
        at_send_command("AT+CGREG=2", NULL); 
        goto __requestRegistrationState_restart;
    }

    if (request == RIL_REQUEST_DATA_REGISTRATION_STATE) {
        const char *gprsStateString[] = {"0", "1"};
        int newgprsState = ((response[0] == 1) || (response[0] == 5));
        if (gprsState != newgprsState) {
            gprsState = newgprsState;
            property_set("net.gprs.gprsState", gprsStateString[gprsState]);
        } 
        if (gprsState && !time_zone_report) {
            ATResponse *p_response = NULL;
            int err = at_send_command_singleline("AT+QLTS", "+QLTS", &p_response);
            if (!(err < 0  || p_response == NULL || p_response->success == 0)) {
                //+QLTS: "13/08/23,06:51:13+32,0"
                char *response;
                char *line = strdup(p_response->p_intermediates->line);
                at_tok_start(&line);
                err = at_tok_nextstr(&line, &response);

                if (err != 0) {
                    LOGE("invalid QLTS line %s\n", p_response->p_intermediates->line);
                } else {
                    RIL_onUnsolicitedResponse (RIL_UNSOL_NITZ_TIME_RECEIVED, response, strlen(response));
                }
                time_zone_report++;
            }
        }
    }
    
    if (count == 3) {
        if ((response[1] == -1) && (response[2] == -1)) {
            //to advoid throw <java.lang.NumberFormatException: Invalid int: "ffffffff"> in GsmServiceStateTracker.java
            count = 1;
        }
    } else if (count > 3) {
        RIL_RadioTechnology radio_tech = RADIO_TECH_UNKNOWN;
        switch (response[3]) {
               case 0: //GSM (Not support on UC20-A)
                    radio_tech = RADIO_TECH_GPRS;
                break;
                case 2: //UTRAN
                    radio_tech = RADIO_TECH_UMTS;
                break;
                case 3: //GSM W/EGPRS
                    radio_tech = RADIO_TECH_EDGE;
                break;
                case 4: //UTRAN W/HSDPA
                    radio_tech = RADIO_TECH_HSDPA;
                break;
                case 5: //UTRAN W/HSUPA
                    radio_tech = RADIO_TECH_HSUPA;
                break;
               case 6: //UTRAN W/HSDPA and HSUPA
                    radio_tech = RADIO_TECH_HSPA;
                break;
               case 7: //LTE
                    radio_tech = RADIO_TECH_LTE;
                break;
                default:
                    radio_tech = RADIO_TECH_UNKNOWN;
                break;
        }
        response[3] = radio_tech;
    }
#endif  

    if (count > 3)
        asprintf(&responseStr[3], "%d", response[3]);

    RIL_onRequestComplete(t, RIL_E_SUCCESS, responseStr, count*sizeof(char*));
    at_response_free(p_response);

    return;
error:
    LOGE("requestRegistrationState must never return an error when radio is on");
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}

static void requestOperator(void *data, size_t datalen, RIL_Token t)
{
    int err;
    int i;
    int skip;
    ATLine *p_cur;
    char *response[3];

#if 1 //quectel
__requestOperator_restart:
#endif
    memset(response, 0, sizeof(response));

    ATResponse *p_response = NULL;

    err = at_send_command_multiline(
        "AT+COPS=3,0;+COPS?;+COPS=3,1;+COPS?;+COPS=3,2;+COPS?",
        "+COPS:", &p_response);

    /* we expect 3 lines here:
     * +COPS: 0,0,"T - Mobile"
     * +COPS: 0,1,"TMO"
     * +COPS: 0,2,"310170"
     */

    if (err != 0) goto error;

    for (i = 0, p_cur = p_response->p_intermediates
            ; p_cur != NULL
            ; p_cur = p_cur->p_next, i++
    ) {
        char *line = p_cur->line;

        err = at_tok_start(&line);
        if (err < 0) goto error;

        err = at_tok_nextint(&line, &skip);
        if (err < 0) goto error;

        // If we're unregistered, we may just get
        // a "+COPS: 0" response
        if (!at_tok_hasmore(&line)) {
            response[i] = NULL;
            continue;
        }

        err = at_tok_nextint(&line, &skip);
        if (err < 0) goto error;

        // a "+COPS: 0, n" response is also possible
        if (!at_tok_hasmore(&line)) {
            response[i] = NULL;
            continue;
        }

        err = at_tok_nextstr(&line, &(response[i]));
        if (err < 0) goto error;
    }

    if (i != 3) {
        /* expect 3 lines exactly */
        goto error;
    }

#if 1 //quectel
    if (!response[0] && !response[1] && !response[2] && (network_debounce_time > 0)) {
        usleep(1*1000*1000);
        network_debounce_time--;
        at_response_free(p_response);
        goto __requestOperator_restart;
    }     
    network_debounce_time = 0;
#endif

    RIL_onRequestComplete(t, RIL_E_SUCCESS, response, sizeof(response));
    at_response_free(p_response);

    return;
error:
    LOGE("requestOperator must not return error when radio is on");
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}

static void requestSendSMS(void *data, size_t datalen, RIL_Token t)
{
    int err;
    const char *smsc;
    const char *pdu;
    int tpLayerLength;
    char *cmd1, *cmd2;
    RIL_SMS_Response response;
    ATResponse *p_response = NULL;

    smsc = ((const char **)data)[0];
    pdu = ((const char **)data)[1];

    tpLayerLength = strlen(pdu)/2;

    // "NULL for default SMSC"
    if (smsc == NULL) {
        smsc= "00";
    }

    asprintf(&cmd1, "AT+CMGS=%d", tpLayerLength);
    asprintf(&cmd2, "%s%s", smsc, pdu);

    err = at_send_command_sms(cmd1, cmd2, "+CMGS:", &p_response);

    if (err != 0 || p_response->success == 0) goto error;

    memset(&response, 0, sizeof(response));

    /* FIXME fill in messageRef and ackPDU */

    RIL_onRequestComplete(t, RIL_E_SUCCESS, &response, sizeof(response));
    at_response_free(p_response);

    return;
error:
    LOGE("%s error", __func__);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}

/*
 * RIL_REQUEST_SEND_SMS_EXPECT_MORE
 */
void requestSendSMSExpectMore(void *data, size_t datalen, RIL_Token t)
{
        (void) data; (void) datalen;

    at_send_command("AT+CMMS=1",NULL);

        requestSendSMS(data, datalen, t);
}

/* wait for a propertyvalue change */
static int wait_for_property(const char *name, const char *desired_value, int maxwait) {
    char value[PROPERTY_VALUE_MAX] = {'\0'};
    int maxnaps = maxwait / 1;

    if (maxnaps < 1) {
        maxnaps = 1;
    }

    while ((maxnaps-- > 0) && (!s_closed)) {
        usleep(1000000);
        if (property_get(name, value, NULL)) {
            LOGI("[%s]: %s %s", __func__, name, value[0] ? value : "\"\"");
            if ((desired_value == NULL) || (strcmp(value, desired_value) == 0)) {
                LOGD("%s ok", __func__);
                return 0;
            }
        } else {
            value[0] = '\0';
            LOGI("[%s]: %s %s", __func__, name, value[0] ? value : "\"\"");
        }
    }

    LOGD("%s fail", __func__);
    return -1; /* failure */
}

static int qsplit(char * line, char * * items) {
    int num = 0, i = 0;

    while (line[i]) {
        while ((line[i] == ' ') || (line[i] == '\t') || (line[i] == '\r') || (line[i] == '\n'))
            i++;

        if (line[i]) {
            items[num++] = &line[i];
            while ((line[i] != ' ') && (line[i] != '\t') && (line[i] != '\0') && (line[i] != '\r') && (line[i] != '\n'))
                i++;
            if ((line[i] == '\0') || (line[i] == '\r') || (line[i] == '\n')) {
                line[i++] = '\0';
                break;
            } else {
                line[i++] = '\0';
            }
        }   
    }

    return num;
}

static int qgetpid(const char *pname) {
    FILE *fp;
    char line[255];
    char * items[20];
    pid_t pid = -1;

    if((fp = popen("ps", "r")) == NULL) {
        LOGE("exec ps error : %d (%s)", errno, strerror(errno));
        return -1;
    }

    if (fgets(line, sizeof(line), fp)) {
        int pid_index, num;
        LOGE("%s", line);
        num = qsplit(line, items);
        for (pid_index = 0; pid_index < num; pid_index++) {
            if (!strcmp("PID", items[pid_index])) {
                break;
            }
        }

        if (pid_index == num) {
            LOGE("canot find PID tag!\n");
            return -1;
        }
                
        while((fgets(line, sizeof(line), fp)) != NULL) {
            char dump[255];
            if (strstr(items[num-1], pname))
                LOGE("%s", line);
            num = qsplit(line, items);
            if (num) {
                char *find = strstr(items[num-1], pname);
                if (find && ((find[-1] == '\0') || (find[-1] == '/'))) {
                    pid =  atoi(items[pid_index]);
                    break;
		}
            }
        }
    }
   
    pclose(fp);
	
    return pid;
}

static int kill_PPP(void)
{
    char pppd_pid_str[PROPERTY_VALUE_MAX];
    int pppd_pid = 0;
    
    LOGD("kill_PPP enter");
    
    pppd_pid = qgetpid("pppd");
    if (pppd_pid > 0) {
        LOGD("send  SIGTERM to pppd %d, and wait ...", pppd_pid);
        if (kill(pppd_pid, SIGTERM) == 0) {
            int waittimes = 5;
            sleep(3);
            while (waittimes-- && (qgetpid("pppd") > 0))
                sleep(2);
            if (!waittimes && (qgetpid("pppd") > 0)) {
                LOGD("send  SIGKILL to pppd %d, and wait ...", pppd_pid);
                kill(pppd_pid, SIGKILL);
                sleep(1);
            }
        }
    }

    // Requires root access...
    property_set("ctl.stop", "quectel-pppd");
    if (wait_for_property("init.svc.quectel-pppd", "stopped", 2) < 0) {
        LOGE("[%s]: stop quectel-pppd failed!", __func__);
        goto error;
    }

    LOGD("%s successful!", __func__);
    return 0;

error:
    LOGD("%s failed!", __func__);
    return -1;
}

//update:Joe.Wang 2014-1-13
static void requestSetupDataCall(void *data, size_t datalen, RIL_Token t)
{
    const char *apn;
    const char *user = NULL;
    const char *pass = NULL;
    const char *pdp_type = NULL;
    char *cmd;
    int err;
    int fd, pppstatus,i;
    FILE *pppconfig;
    size_t cur = 0;
    ssize_t written, rlen;
    char status[32] = {0};
    char *buffer;
    long buffSize, len;
    int retry = 10;
    ATResponse *p_response = NULL;
    int iRetry = PPP_START_RETRY_TIME;

    int n = 1;
    RIL_Data_Call_Response_v6 responses;
    char ppp_dnses[(PROPERTY_VALUE_MAX * 2) + 3] = {'\0'};
    char ppp_local_ip[PROPERTY_VALUE_MAX] = {'\0'};
    char ppp_dns1[PROPERTY_VALUE_MAX] = {'\0'};
    char ppp_dns2[PROPERTY_VALUE_MAX] = {'\0'};
    char ppp_gw[PROPERTY_VALUE_MAX] = {'\0'};
    char exit_code[PROPERTY_VALUE_MAX] = {'\0'};
       
    apn = ((const char **)data)[2];
    user = ((char **)data)[3];
    if(user != NULL)
    {
        if (strlen(user)<2){
            LOGD("[%s]: length of user name is too short", __func__);
            user = "dummy";
        }else{
            LOGD("[%s]: Input user name is %s", __func__, user);
        }
    } else{
        LOGD("[%s]: user name is null", __func__);
        user = "dummy";
    }

    pass = ((char **)data)[4];
    if(pass != NULL)
    {
        if (strlen(pass)<2){
            LOGD("[%s]: length of password is too short\r\n", __func__);
            pass = "dummy";
        }else{
            LOGD("[%s]: Input password is %s\r\n", __func__,pass);
        }
    } else{
        LOGD("[%s]: pass word is null\r\n", __func__);
        pass = "dummy";
    }
    LOGI("*************************************");
    LOGI("USER:%s",user);
    LOGI("PASS:%s",pass);
    LOGI("*************************************");
    property_set("net.gprs.user", user);
    property_set("net.gprs.password", pass);
    property_set("net.ppp0.user", user);
    property_set("net.ppp0.password", pass);

#if 1 // quectel Set DTR Function Mode 
//ON->OFF on DTR: Disconnect data call, change to command mode. During state DTR = OFF, auto-answer function is disabled
    at_send_command("AT&D2", NULL);
#endif

     //Make sure there is no existing connection or pppd instance running
    if(kill_PPP() < 0) {
        goto error;
    }

    if (gprsState == 0) {
        LOGE("gprsState is out of service!");
        goto error;
    }
    
    LOGD("requesting data connection to APN '%s'!\n", apn);

    if (datalen > 6 * sizeof(char *)) {
        pdp_type = ((const char **)data)[6];
    } else {
        pdp_type = "IP";
    }
     
    asprintf(&cmd, "AT+CGDCONT=1,\"%s\",\"%s\"", pdp_type, apn);

    //FIXME check for error here
    err = at_send_command(cmd, NULL);
    free(cmd);

    sleep(2); //Wait for the modem to finish
	
    /* start the gprs pppd */
    property_set(PPPD_EXIT_CODE, "");
    property_set("net.ppp0.local-ip", "");
    property_set("ctl.start", "quectel-pppd");

    sleep(3); // sleep to wait pppd finish

        
    if (wait_for_property("net.ppp0.local-ip", NULL, PPP_START_RETRY_TIME * 10) < 0) {
        LOGE("[%s]: fail to get IP\r\n", __func__);
        goto error;
    }

    property_get("net.ppp0.local-ip", ppp_local_ip, NULL);
    property_get("net.ppp0.dns1", ppp_dns1, NULL);
    property_get("net.ppp0.dns2", ppp_dns2, NULL);
    property_get("net.ppp0.gw", ppp_gw, NULL);
    sprintf(ppp_dnses, "%s %s", ppp_dns1, ppp_dns2);

    LOGI("Got net.ppp0.local-ip: %s\n", ppp_local_ip);

    responses.status = 0;
    responses.suggestedRetryTime = -1;
    responses.cid = 1;
    responses.active = 2;
    responses.type = (char*)"PPP";
    responses.ifname = (char*)PPP_TTY_PATH;
    responses.addresses = ppp_local_ip;
    responses.dnses = ppp_dnses;
    responses.gateways = ppp_gw;

    RIL_onRequestComplete(t, RIL_E_SUCCESS, &responses, sizeof(RIL_Data_Call_Response_v6));
    at_response_free(p_response);

#if 1 // quectel Set DTR Function Mode 
    if (USB_HOST_SUSPEND_SUPPORT == 0) //if the usb host donot support suspend
        at_send_command("AT&D0", NULL); //TA ignores status on DTR, will donot disconnect ppp when DTR is assert
#endif
    return;

error:
    kill_PPP();
    LOGE("Unable to setup PDP in %s\n", __func__);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}

static void requestSMSAcknowledge(void *data, size_t datalen, RIL_Token t)
{
    int ackSuccess;
    int err;

    ackSuccess = ((int *)data)[0];

    if (ackSuccess == 1) {
        err = at_send_command("AT+CNMA=1", NULL);
    } else if (ackSuccess == 0)  {
        err = at_send_command("AT+CNMA=2", NULL);
    } else {
        LOGE("unsupported arg to RIL_REQUEST_SMS_ACKNOWLEDGE\n");
        goto error;
    }

    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    return;
error:
    LOGE("%s error", __func__);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);

}

static void  requestSIM_IO(void *data, size_t datalen, RIL_Token t)
{
    ATResponse *p_response = NULL;
    RIL_SIM_IO_Response sr;
    int err;
    char *cmd = NULL;
    RIL_SIM_IO_v6 *p_args;
    char *line;

    memset(&sr, 0, sizeof(sr));

    p_args = (RIL_SIM_IO_v6 *)data;

    /* FIXME handle pin2 */

    if (p_args->data == NULL) {
        LOGI("[%s]: p_args->data is NULL\r\n", __func__);
        asprintf(&cmd, "AT+CRSM=%d,%d,%d,%d,%d",
                    p_args->command, p_args->fileid,
                    p_args->p1, p_args->p2, p_args->p3);
    } else {
#if 1 //quectel
        asprintf(&cmd, "AT+CRSM=%d,%d,%d,%d,%d,\"%s\"",
#else
        asprintf(&cmd, "AT+CRSM=%d,%d,%d,%d,%d,%s",
#endif
                    p_args->command, p_args->fileid,
                    p_args->p1, p_args->p2, p_args->p3, p_args->data);
    }
    
    err = at_send_command_singleline(cmd, "+CRSM:", &p_response);

    if (err < 0 || p_response->success == 0) {
        LOGI("[%s]: send +CRSM error\r\n", __func__);
        goto error;
    }

    line = p_response->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &(sr.sw1));
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &(sr.sw2));
    if (err < 0) goto error;

    if (at_tok_hasmore(&line)) {
        err = at_tok_nextstr(&line, &(sr.simResponse));
        if (err < 0) goto error;
    }

    LOGD("[%s]: RIL_SIM_IO_Response Complete sr.sw1=%d, sr.sw2=%d, sr.simResponse=%s\r\n",
        __func__, sr.sw1, sr.sw2, sr.simResponse);
    
    RIL_onRequestComplete(t, RIL_E_SUCCESS, &sr, sizeof(sr));
    at_response_free(p_response);
    free(cmd);

    return;
error:
    LOGE("%s error\n", __func__);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
    free(cmd);

}

static int getRemainingTimes(int request)
{
    ATResponse *p_response = NULL;
    ATLine *p_cur = NULL;
    char *line = NULL;
    int err;
    //int n = 0;
    char *fac = NULL;
    int pin1_remaining_times; 
    int puk1_remaining_times;
    int pin2_remaining_times;
    int puk2_remaining_times;

    err = at_send_command_multiline("AT+QPINC?", "+QPINC", &p_response);
    //for (p_cur = p_response->p_intermediates; p_cur != NULL;
    //     p_cur = p_cur->p_next)
    //    n++;

    if(err < 0  || p_response == NULL || p_response->success == 0) goto error;
    
    for (p_cur = p_response->p_intermediates; p_cur != NULL;p_cur = p_cur->p_next) 
    {
        char *line = p_cur->line;

        at_tok_start(&line);

        err = at_tok_nextstr(&line,&fac);
        if(err < 0) goto error;

        if(!strncmp(fac,"SC",2))
        {
            err = at_tok_nextint(&line,&pin1_remaining_times);
            if(err < 0) goto error;
            err = at_tok_nextint(&line,&puk1_remaining_times);
            if(err < 0) goto error;
        }
        else if(!strncmp(fac,"P2",2))
        {
            err = at_tok_nextint(&line,&pin2_remaining_times);
            if(err < 0) goto error;
            err = at_tok_nextint(&line,&puk2_remaining_times);
            if(err < 0) goto error;
        }
        else
            goto error;
    }
    LOGI("PIN1:%d\nPIN2:%d\nPUN1:%d\nPUN2:%d",pin1_remaining_times,pin2_remaining_times,puk1_remaining_times,puk2_remaining_times);

    free(p_response);
    free(p_cur);
#if 0
    if(!strcmp(pinType,"PIN1")) return pin1_remaining_times;
    else  if(!strcmp(pinType,"PIN2")) return pin2_remaining_times; 
    else  if(!strcmp(pinType,"PUK1")) return puk1_remaining_times; 
    else  if(!strcmp(pinType,"PUK2")) return puk2_remaining_times; 
    else return -1;
#else
    switch(request)
    {
        case RIL_REQUEST_SET_FACILITY_LOCK:
        case RIL_REQUEST_ENTER_SIM_PIN:
        case RIL_REQUEST_CHANGE_SIM_PIN:
            return pin1_remaining_times;
        case RIL_REQUEST_ENTER_SIM_PIN2:
        case RIL_REQUEST_CHANGE_SIM_PIN2:
            return pin2_remaining_times;
        case RIL_REQUEST_ENTER_SIM_PUK:
            return puk1_remaining_times;
        case RIL_REQUEST_ENTER_SIM_PUK2:
            return puk2_remaining_times;
        default:
            return -1;
    }
#endif
error:
    return -1;
}

static void  requestQuestFacility(void*  data, size_t  datalen, RIL_Token  t)
{
    ATResponse   *p_response = NULL;
    int           err;
    char*         cmd = NULL;
    int status;
    char *fac = NULL;
    char *mode = "2";
    char *password = NULL;
    char *class = NULL;
    char *line = NULL;

    LOGE("%s Enter\n", __func__);

    fac = ((char **)data)[0];
    password = ((char **)data)[1];
    class = ((char **)data)[2];

    asprintf(&cmd,"AT+CLCK=\"%s\",%s,\"%s\",%s",fac,mode,password,class);
    err = at_send_command_singleline(cmd, "+CLCK:", &p_response);
    if(err < 0  || p_response == NULL || p_response->success == 0) goto error;

    line = p_response->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &status);
    if (err < 0) goto error;

    RIL_onRequestComplete(t, RIL_E_SUCCESS, &status, sizeof(int *));
    goto out;

error:
    LOGE("%s error\n", __func__);
    RIL_onRequestComplete(t, RIL_E_PASSWORD_INCORRECT, NULL, 0);
    //RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
out:

    LOGE("%s leave\n", __func__);

    free(cmd);
    at_response_free(p_response);
}

static void  requestSetFacility(void*  data, size_t  datalen, RIL_Token  t,int request)
{
    ATResponse   *p_response = NULL;
    int           err;
    char*         cmd = NULL;
    int remaining_times = -1;
    char *fac = NULL;
    char *mode = NULL;
    char *password = NULL;
    char *class = NULL;

    LOGE("%s Enter\n", __func__);

    fac = ((char **)data)[0];
    mode = ((char **)data)[1];
    password = ((char **)data)[2];
    class = ((char **)data)[3];

    asprintf(&cmd,"AT+CLCK=\"%s\",%s,\"%s\",%s",fac,mode,password,class);
    err = at_send_command(cmd,&p_response);
    if(err < 0 || p_response == NULL || p_response->success == 0) goto error;
    
    remaining_times = getRemainingTimes(request);
    LOGI("remaining_times = %d",remaining_times);
    RIL_onRequestComplete(t, RIL_E_SUCCESS, &remaining_times, sizeof(int *));
    goto out;

error:
    LOGE("%s error\n", __func__);
    RIL_onRequestComplete(t, RIL_E_PASSWORD_INCORRECT, NULL, 0);
    //RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
out:
    LOGE("%s leave\n", __func__);
    free(cmd);
}

static void  requestChangeSimPin(void*  data, size_t  datalen, char *fac, RIL_Token  t,int request)
{
    ATResponse   *p_response = NULL;
    int           err;
    char*         cmd = NULL;
    int remaining_times = -1;
    char *oldpin = NULL,*newpin = NULL;

    LOGE("%s Enter\n", __func__);

    oldpin = ((char **)data)[0];
    newpin = ((char **)data)[1];

    asprintf(&cmd,"AT+CPWD=\"%s\",\"%s\",\"%s\"",fac,oldpin,newpin);
    err = at_send_command(cmd,&p_response);
    if(err < 0 || p_response == NULL || p_response->success == 0) goto error;
    
    remaining_times = getRemainingTimes(request);
    LOGI("remaining_times = %d",remaining_times);
    RIL_onRequestComplete(t, RIL_E_SUCCESS, &remaining_times, sizeof(int *));
    goto out;

error:
    LOGE("%s error\n", __func__);
    RIL_onRequestComplete(t, RIL_E_PASSWORD_INCORRECT, NULL, 0);
    //RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
out:
    LOGE("%s leave\n", __func__);
    free(cmd);
}

static void  requestEnterSimPin(void*  data, size_t  datalen, RIL_Token  t,int request)
{
    ATResponse   *p_response = NULL;
    int           err;
    char*         cmd = NULL;
    const char**  strings = (const char**)data;
    int remaining_times = -1;

    LOGE("%s Enter\n", __func__);

    if (( datalen == sizeof(char*)) || !strings[1] || !strings[1][0]) {
        asprintf(&cmd, "AT+CPIN=%s", strings[0]);
    } else {
        asprintf(&cmd, "AT+CPIN=\"%s\",\"%s\"", strings[0], strings[1]);
    }

    err = at_send_command(cmd, &p_response);
    if(err < 0 || p_response == NULL || p_response->success == 0) goto error;

    remaining_times = getRemainingTimes(request);
    LOGI("remaining_times = %d",remaining_times);
    RIL_onRequestComplete(t, RIL_E_SUCCESS, &remaining_times, sizeof(int *));
    goto out;

error:
    LOGE("%s error\n", __func__);
    RIL_onRequestComplete(t, RIL_E_PASSWORD_INCORRECT, NULL, 0);
    //RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
out:
    LOGE("%s leave\n", __func__);
    free(cmd);
    at_response_free(p_response);
}


static void  requestSendUSSD(void *data, size_t datalen, RIL_Token t)
{
    ATResponse   *p_response = NULL;
    int err;
    char *cmd = NULL;
    const char *ussdRequest;
    int mode,dcs;

#ifdef CUSD_USE_UCS2_MODE
	int i = 0;
	char ucs2[100];
	ussdRequest = (char *)(data);
	while (ussdRequest[i]) {
		sprintf(ucs2+i*4, "%04X", ussdRequest[i]);
		i++;
	}
	data = ucs2;
#endif
    ussdRequest = (char *)(data);
    mode = 1;
//    dcs = 15;

    //asprintf(&cmd,"AT+CUSD=%d,%s,%d",mode,ussdRequest,dcs);                                                                                              
    asprintf(&cmd,"AT+CUSD=%d,\"%s\"",mode,ussdRequest);                                                                                              
    err = at_send_command(cmd,&p_response);
    if(err != 0 || p_response == NULL || p_response->success == 0)
    {
        goto error;
    }

#if 1 //quectel
    ussd_pending_index++;
    RIL_requestTimedCallback(onUssdTimedCallback, (void *)ussd_pending_index, &ussd_timeout_timeval);
#endif
    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    return;
error:
    LOGE("[Joe]ERROR: %s failed",__func__);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
// @@@ TODO

}

/**
 *  Wythe:Add on 2013-04-02 for 4.0 ril
 *  Add new function to handle more requests.
 *  Must response to RIL.JAVA.
 */

/*
* Function: requestScreenState
* Purpose : handle the request when screen is on or off.
* Request : RIL_REQUEST_SCREEN_STATE
*/
static void requestScreenState(void *data, size_t datalen, RIL_Token t)
{
#if 1 //quectel
    if ((USB_HOST_SUSPEND_SUPPORT == 0) && (ql_is_UC20)) { //if the usb host donot support suspend
        if (((int *)data)[0]) {
            at_send_command("AT+QCFG=\"urc/ri/smsincoming\",\"pulse\",120", NULL);
            at_send_command("AT+QCFG=\"urc/delay\",0", NULL);
        } else {
            at_send_command("AT+QCFG=\"urc/ri/smsincoming\",\"pulse\",2000", NULL);
            at_send_command("AT+QCFG=\"urc/delay\",1", NULL);
        }
    }
#endif

    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}

/*
*   Function: requestBaseBandVersion
*   Purpose : return string of BaseBand version
*   Request : RIL_REQUEST_BASEBAND_VERSION
*/
static void requestBaseBandVersion(void *data, size_t datalen, RIL_Token t)
{
    int err;
    ATResponse *atResponse = NULL;
    char *line;

    err = at_send_command_singleline("AT+CGMR", "\0", &atResponse);

    if(err != 0){
        LOGE("[Wythe]%s() Error Reading Base Band Version!",__func__);
        RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
        return;
    }

    line = atResponse->p_intermediates->line;

    RIL_onRequestComplete(t, RIL_E_SUCCESS, line, sizeof(char *));
    
    at_response_free(atResponse);
}


/*
*   Function: requestGetIMEISV 
*   Purpose : return the IMEI SV(software version)
*   Request : RIL_REQUEST_GET_IMEISV
*/
static void requestGetIMEISV(RIL_Token t)
{
    RIL_onRequestComplete(t, RIL_E_SUCCESS, (void *)00, sizeof(char *));
}

/**
 *  Function: requestSetPreferredNetworkType
 *  Purpose : Requests to set the preferred network type for searching and registering
 *            (CS/PS domain, RAT, and operation mode).
 *  Request : RIL_REQUEST_SET_PREFERRED_NETWORK_TYPE
 */ 
static void requestSetPreferredNetworkType(void *data, size_t datalen, RIL_Token t)
{
	int rat,err;
    const char* nwscanmode;
    const char* nwscanseq = "AT+QCFG=\"nwscanseq\",0"; /* AUTO */
	
	assert (datalen >= sizeof(int *));
	rat = ((int *)data)[0];

	/*  specific 	
     * ((int *)data)[0] is == 0 for GSM/WCDMA (WCDMA preferred)
     * ((int *)data)[0] is == 1 for GSM only
     * ((int *)data)[0] is == 2 for WCDMA only
     * ((int *)data)[0] is == 3 for GSM/WCDMA (auto mode, according to PRL)
     * ((int *)data)[0] is == 4 for CDMA and EvDo (auto mode, according to PRL)
     * ((int *)data)[0] is == 5 for CDMA only
     * ((int *)data)[0] is == 6 for EvDo only
     * ((int *)data)[0] is == 7 for GSM/WCDMA, CDMA, and EvDo (auto mode, according to PRL)
	 */

     /**
      * Add to handle request for preferred network type setting
      * Wythe 2013-9-27
      */
      
    /* in hardware\ril\include\ril.h
     * PREF_NET_TYPE_GSM_WCDMA                = 0, GSM/WCDMA (WCDMA preferred)
     * PREF_NET_TYPE_GSM_ONLY                 = 1, GSM only 
     * PREF_NET_TYPE_WCDMA                    = 2, WCDMA
     * PREF_NET_TYPE_GSM_WCDMA_AUTO           = 3, GSM/WCDMA (auto mode, according to PRL)
     * PREF_NET_TYPE_CDMA_EVDO_AUTO           = 4, CDMA and EvDo (auto mode, according to PRL)
     * PREF_NET_TYPE_CDMA_ONLY                = 5, CDMA only
     * PREF_NET_TYPE_EVDO_ONLY                = 6, EvDo only
     * PREF_NET_TYPE_GSM_WCDMA_CDMA_EVDO_AUTO = 7, GSM/WCDMA, CDMA, and EvDo (auto mode, according to PRL)
     * PREF_NET_TYPE_LTE_CDMA_EVDO            = 8, LTE, CDMA and EvDo
     * PREF_NET_TYPE_LTE_GSM_WCDMA            = 9, LTE, GSM/WCDMA
     * PREF_NET_TYPE_LTE_CMDA_EVDO_GSM_WCDMA  = 10, LTE, CDMA, EvDo, GSM/WCDMA
     * PREF_NET_TYPE_LTE_ONLY                 = 11  LTE only
     */
	
	switch (rat) {
        case PREF_NET_TYPE_GSM_ONLY:    /* GSM ONLY*/
            nwscanmode = "AT+QCFG=\"nwscanmode\",1";
            break;
        case PREF_NET_TYPE_WCDMA: /* WCDMA ONLY*/
            nwscanmode = "AT+QCFG=\"nwscanmode\",2";
            break;
           
        case PREF_NET_TYPE_GSM_WCDMA: /* GSM/WCDMA (WCDMA preferred) */
        default:
            nwscanmode = "AT+QCFG=\"nwscanmode\",0";
            nwscanseq = "AT+QCFG=\"nwscanseq\",2"; /* WCDMA prior to GSM */
			break; 
	}

	/* Set mode */
    err = at_send_command(nwscanseq, NULL);
        if (err != 0) 
            goto error;
    err = at_send_command(nwscanmode, NULL);
        if (err != 0) 
		goto error;

	RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, sizeof(int));
	return;
	
error:
	LOGE("ERROR: requestSetPreferredNetworkType() failed\n");
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

/*
*   Function: requestReportSTKServiceIsRunning
*   Purpose :
*   Request : RIL_REQUEST_REPORT_STK_SERVICE_IS_RUNNING
*/
static void requestReportSTKServiceIsRunning(RIL_Token t)
{
    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}

/* 
 * Function: requestDeactivateDefaultPDP
 * Purpose :
 * Request : RIL_REQUEST_DEACTIVATE_DATA_CALL
 */ 
static void requestDeactivateDefaultPDP(void *data, size_t datalen, RIL_Token t)
{
    char * cid;
    char *cmd;

    LOGD("requestDeactivateDefaultPDP()");

    kill_PPP();
    cid = ((char **)data)[0];
    asprintf(&cmd, "AT+CGACT=0,%s", cid);
    at_send_command(cmd, NULL);
    free(cmd);
    
    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    return;

error:
    LOGE("%s error\n",__func__);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

/**
 * Funtcion:requestLastPDPFailCause
 *
 * Purpose :Requests the failure cause code for the most recently failed PDP
 * context activate.
 *
 * Request : RIL_REQUEST_LAST_CALL_FAIL_CAUSE.
 *
 */
static int s_lastPdpFailCause = PDP_FAIL_ERROR_UNSPECIFIED;
static void requestLastPDPFailCause(RIL_Token t)
{
    RIL_onRequestComplete(t, RIL_E_SUCCESS, &s_lastPdpFailCause, sizeof(int));
} 

static void requestGsmGetBroadcastSmsConfig(RIL_Token t)
{
	RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);
}

static void requestGsmSetBroadcastSmsConfig(void *data, size_t datalen, RIL_Token t)
{
	RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);
}

static void requestGsmSmsBroadcastActivation(void *data, size_t datalen, RIL_Token t)
{
	RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);
}

static void requestDtmfStop(RIL_Token t)
{
	RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);
}

static void requestQueryCallForwardStatus(void *data, size_t datalen, RIL_Token t)
{
    RIL_CallForwardInfo *callForwardInfo = (RIL_CallForwardInfo *)data;
    RIL_CallForwardInfo *r_callForwardInfo[10];
    char *cmd = NULL;
    int err;
    ATLine *tmp = NULL;
    char *line = NULL;
    ATResponse *atResponse;
    int i = 0;

    asprintf(&cmd,"AT+CCFC=%d,2",callForwardInfo->reason);
    err = at_send_command_multiline(cmd,"+CCFC:",&atResponse);
    free(cmd);
    if (err < 0 || atResponse->success == 0)
        goto error;

    tmp = atResponse->p_intermediates;
    while (tmp) 
    {
        r_callForwardInfo[i] = (RIL_CallForwardInfo *)alloca(sizeof(RIL_CallForwardInfo));
        bzero(r_callForwardInfo[i],sizeof(RIL_CallForwardInfo));
        r_callForwardInfo[i]->number = (char *)alloca(15);
        bzero(r_callForwardInfo[i]->number,15);
        line = tmp->line;
        err = at_tok_start(&line);
        if (err < 0) goto error;
        err = at_tok_nextint(&line,&(r_callForwardInfo[i]->status));
        if (err < 0) goto error;
        if(r_callForwardInfo[i]->status != 0)
        {
            err = at_tok_nextint(&line,&(r_callForwardInfo[i]->serviceClass));
            if (err < 0) goto error;
            err = at_tok_nextstr(&line,&(r_callForwardInfo[i]->number));
            if (err < 0) goto error;
            LOGI("status:%d\nserviceClass:%d\nnumber:%s\n",r_callForwardInfo[i]->status,r_callForwardInfo[i]->serviceClass,r_callForwardInfo[i]->number);
        }
        i++;

        tmp = tmp->p_next;
    }

	RIL_onRequestComplete(t, RIL_E_SUCCESS, r_callForwardInfo, i * sizeof(RIL_CallForwardInfo *));
    return;
error:
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void requestSetCallForward(void *data, size_t datalen, RIL_Token t)
{
    char *cmd = NULL;
    int err;
    RIL_CallForwardInfo *callForwardInfo = (RIL_CallForwardInfo *)data;

    if(callForwardInfo->serviceClass == 0)
    {
        asprintf(&cmd,"AT+CCFC=%d,%d,\"%s\",%d",
                 callForwardInfo->reason,callForwardInfo->status,
                 callForwardInfo->number,callForwardInfo->toa);
    }
    else
    {
        asprintf(&cmd,"AT+CCFC=%d,%d,\"%s\",%d,%d",
                 callForwardInfo->reason,callForwardInfo->status,
                 callForwardInfo->number,callForwardInfo->toa,callForwardInfo->serviceClass);
    }

    err = at_send_command(cmd, NULL);
    free(cmd);
    if(err != 0) goto error;

	RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    return;
error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void requestGetClir(RIL_Token t)
{
    int err;
    int response[2];
    int n; 
    int m;
    char *line = NULL;
    ATResponse *atResponse = NULL;

    err = at_send_command_singleline("AT+CLIR?","+CLIR:",&atResponse);
	if (err != 0) goto error;
    
	line = atResponse->p_intermediates->line;
	err = at_tok_start(&line);
	if (err < 0) goto error;
    
    err = at_tok_nextint(&line,&n);
	if (err < 0) goto error;

    err = at_tok_nextint(&line,&m);
	if (err < 0) goto error;

    response[0] = n;
    response[1] = m;

	RIL_onRequestComplete(t, RIL_E_SUCCESS, response, sizeof(response));
    return ;

error:
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void requestSetClir(void *data, size_t datalen, RIL_Token t)
{
    int err;
    char *cmd = NULL;

    asprintf(&cmd, "AT+CLIR=%d",((int *)data)[0]);
    err = at_send_command(cmd,NULL);
	if (err != 0) goto error;

	RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    free(cmd);
    return ;

error:
    free(cmd);
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void requestQueryCallWaiting(void *data, size_t datalen, RIL_Token t)
{
    int err;
    char *cmd = NULL;
    int response[2];
    int status; 
    int class;
    char *line = NULL;
    ATResponse *atResponse = NULL;

    asprintf(&cmd, "AT+CCWA=1,2,%d",((int *)data)[0]);
    err = at_send_command_singleline(cmd,"+CCWA:",&atResponse);
	if (err != 0) goto error;
    
	line = atResponse->p_intermediates->line;
	err = at_tok_start(&line);
	if (err < 0) goto error;
    
    err = at_tok_nextint(&line,&status);
	if (err < 0) goto error;

    err = at_tok_nextint(&line,&class);
	if (err < 0) goto error;

    response[0] = status;
    response[1] = class;

	RIL_onRequestComplete(t, RIL_E_SUCCESS, response, sizeof(response));
    free(cmd);
    return ;

error:
    free(cmd);
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void requestSetCallWaiting(void *data, size_t datalen, RIL_Token t)
{
    int err;
    char *cmd = NULL;

    asprintf(&cmd, "AT+CCWA=1,%d,%d",((int *)data)[0],((int *)data)[1]);
    err = at_send_command(cmd,NULL);
	if (err != 0) goto error;

	RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    free(cmd);
    return ;

error:
    free(cmd);
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void requestDtmfStart(void *data, size_t datalen, RIL_Token t)
{
    int err;
    char *cmd;

    asprintf(&cmd,"AT+VTS=%c",((char *)data)[0]);

    err = at_send_command(cmd,NULL);
	if (err != 0) 
		goto error;

	RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    return ;
error:
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void requestSetMute(void *data, size_t datalen, RIL_Token t)
{
	int err;
    char *cmd = NULL;
    
	assert (datalen >= sizeof(int *));

	/* mute */
    asprintf(&cmd, "AT+CMUT=%d", ((int*)data)[0]);
	err = at_send_command(cmd, NULL);
    
	if (err != 0) 
		goto error;

	RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
	return;

error:
	LOGE("ERROR: requestSetMute failed");
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);

}

static void requestGetMute(RIL_Token t)
{
	int err;
	ATResponse *atResponse = NULL;
	int response[1];
	char *line;

	err = at_send_command_singleline("AT+CMUT?", "+CMUT:", &atResponse);

	if (err != 0)
		goto error;

	line = atResponse->p_intermediates->line;

	err = at_tok_start(&line);
	if (err < 0) goto error;

	err = at_tok_nextint(&line, &response[0]);
	if (err < 0) goto error;

	RIL_onRequestComplete(t, RIL_E_SUCCESS, response, sizeof(response));
	at_response_free(atResponse);

	return;

error:
	LOGE("ERROR: requestGetMute failed");
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	at_response_free(atResponse);
}

/**
 * Funtcion:requestResetRadio
 *
 * Purpose :reset module for JAVA
 *
 * Request : RIL_REQUEST_RESET_RADIO.
 *
 */
static void requestResetRadio(RIL_Token t)
{
	int err = 0;

	/* Reset MS */
	err = at_send_command("AT+QRST=1,0", NULL);
	if(err != 0)
		goto error;

    /* sleep 5s to wait for reboot */
    sleep(5);
		
	RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
	return;
	
error:
    LOGE("[%s]: error for qrst\n", __func__);
	RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);
	return;
}

//Add by wythe on 2013-9-27 ->start

/**
 * Function:request get preferred network type
 *
 * Purpose :Query the preferred network type (CS/PS domain, RAT, and operation mode)
 * for searching and registering.
 *
 * Request :RIL_REQUEST_GET_PREFERRED_NETWORK_TYPE
 */ 
static void requestGetPreferredNetworkType(RIL_Token t)
{
	/*
	 * AT+QCFG="nwscanmode",0       AUTO,WCDMA preferred
	 * AT+QCFG="nwscanmode",1       GSM ONLY
	 * AT+QCFG="nwscanmode",2       WCDMA ONLY
	 */

	ATResponse *atResponse = NULL;
	int err;
	char *line;
	int ret;
	int response = PREF_NET_TYPE_GSM_WCDMA;

	err = at_send_command_singleline("AT+QCFG=\"nwscanmode\"", "+QCFG:", &atResponse);

	if (err != 0) {
		// assume radio is off
		goto error;
	}

	line = atResponse->p_intermediates->line;

	err = at_tok_start(&line);
	if (err < 0) goto error;

	err = skipComma(&line);
    if (err < 0) goto error;

	err = at_tok_nextint(&line, &ret);
	if (err < 0) goto error;

	/* Based on reported +QCFG: "nwscanmode" */
	if (ret == 1) {
		response = PREF_NET_TYPE_GSM_ONLY;  /* GSM only */
	} else if (ret == 2) {
		response = PREF_NET_TYPE_WCDMA; /* WCDMA only */
	} else {
		response = PREF_NET_TYPE_GSM_WCDMA; /* for 3G Preferred */
	}
	
	D("requestGetPreferredNetworkType() mode:%d\n",response);
	RIL_onRequestComplete(t, RIL_E_SUCCESS, &response, sizeof(int));
	at_response_free(atResponse);	
	return;
	
error:
	LOGE("ERROR: requestGetPreferredNetworkType() failed - modem does not support command\n");
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	at_response_free(atResponse);
}
//Add by wythe on 2013-9-27 ->end

struct operatorPollParams {
    RIL_Token t;
    int loopcount;
}; 


#define REPOLL_OPERATOR_SELECTED 30     
static const struct timeval TIMEVAL_OPERATOR_SELECT_POLL = { 2, 0 }; 

/**
 * Poll +COPS? and return a success, or if the loop counter reaches
 * REPOLL_OPERATOR_SELECTED, return generic failure.
 */
static void pollOperatorSelected(void *params)
{
    int err = 0;
    int response = 0;
    char *line = NULL;
    ATResponse *atResponse = NULL;
    struct operatorPollParams *poll_params;
    RIL_Token t;

    assert(params != NULL);

    poll_params = (struct operatorPollParams *) params;
    t = poll_params->t;

    if (poll_params->loopcount >= REPOLL_OPERATOR_SELECTED)
        goto error;

    err = at_send_command_singleline("AT+COPS?", "+COPS:", &atResponse);
    if (err != 0)
        goto error;

    line = atResponse->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0)
        goto error;

    err = at_tok_nextint(&line, &response);
    if (err < 0)
        goto error;

    /* If we don't get more than the COPS: {0-4} we are not registered.
       Loop and try again. */
    if (!at_tok_hasmore(&line)) {
            poll_params->loopcount++;
            RIL_requestTimedCallback(pollOperatorSelected,
                            poll_params, &TIMEVAL_OPERATOR_SELECT_POLL);
    } else {
        /* We got operator, throw a success! */
        RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
        free(poll_params);
    }

    at_response_free(atResponse);
    return;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    free(poll_params);
    at_response_free(atResponse);
    return;
}

/**
 *  Add handler for RIL_REQUEST_SET_NETWORK_SELECTION_AUTOMATIC
 *  wythe add on 2014-3-28
 */
void requestSetNetworkSelectionAutomatic(RIL_Token t)
{
    int err = 0;
    ATResponse *atResponse = NULL;
    int mode = 0;
    int skip;
    char *line;
    char *netOperator = NULL;
    struct operatorPollParams *poll_params = NULL;

    poll_params = (struct operatorPollParams*)
					malloc(sizeof(struct operatorPollParams));
    if (NULL == poll_params)
        goto error;

    /* First check if we are already scanning or in manual mode */
    err = at_send_command_singleline("AT+COPS=3,2;+COPS?", "+COPS:", &atResponse);
    if (err != 0)
        goto error;

    line = atResponse->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0)
        goto error;

    /* Read network selection mode */
    err = at_tok_nextint(&line, &mode);
    if (err < 0)
        goto error;

    /* If we're unregistered, we may just get
       a "+COPS: 0" response. */
    if (!at_tok_hasmore(&line)) {
        if (mode == 1) {
            LOGD("%s() Changing manual to automatic network mode", __func__);
            goto do_auto;
        } else
            goto check_reg;
    }

    err = at_tok_nextint(&line, &skip);
    if (err < 0)
        goto error;

    /* A "+COPS: 0, n" response is also possible. */
    if (!at_tok_hasmore(&line)) {
        if (mode == 1) {
            LOGD("%s() Changing manual to automatic network mode", __func__);
            goto do_auto;
        } else
            goto check_reg;
    }

    /* Read numeric operator */
    err = at_tok_nextstr(&line, &netOperator);
    if (err < 0)
        goto error;

    /* If operator is found then do a new scan,
       else let it continue the already pending scan */
    if (netOperator && strlen(netOperator) == 0) {
        if (mode == 1) {
            LOGD("%s() Changing manual to automatic network mode", __func__);
            goto do_auto;
        } else
            goto check_reg;
    }

    /* Operator found */
    if (mode == 1) {
        LOGD("%s() Changing manual to automatic network mode", __func__);
        goto do_auto;
    } else {
        LOGD("%s() Already in automatic mode with known operator, trigger a new network scan",
	    __func__);
        goto do_auto;
    }

    /* Check if module is scanning,
       if not then trigger a rescan */
check_reg:
    at_response_free(atResponse);
    atResponse = NULL;

    /* Check CS domain first */
    err = at_send_command_singleline("AT+CREG?", "+CREG:", &atResponse);
    if (err != 0)
        goto error;

    line = atResponse->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0)
        goto error;

    /* Read registration unsolicited mode */
    err = at_tok_nextint(&line, &mode);
    if (err < 0)
        goto error;

    /* Read registration status */
    err = at_tok_nextint(&line, &mode);
    if (err < 0)
        goto error;

    /* If scanning has stopped, then perform a new scan */
    if (mode == 0) {
        LOGD("%s() Already in automatic mode, but not currently scanning on CS,"
	     "trigger a new network scan", __func__);
        goto do_auto;
    }

    /* Now check PS domain */
    at_response_free(atResponse);
    atResponse = NULL;
    err = at_send_command_singleline("AT+CGREG?", "+CGREG:", &atResponse);
    if (err != 0)
        goto error;

    line = atResponse->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0)
        goto error;

    /* Read registration unsolicited mode */
    err = at_tok_nextint(&line, &mode);
    if (err < 0)
        goto error;

    /* Read registration status */
    err = at_tok_nextint(&line, &mode);
    if (err < 0)
        goto error;

    /* If scanning has stopped, then perform a new scan */
    if (mode == 0) {
        LOGD("%s() Already in automatic mode, but not currently scanning on PS,"
	     "trigger a new network scan", __func__);
        goto do_auto;
    }
    else
    {
        LOGD("%s() Already in automatic mode and scanning", __func__);
        goto finish_scan;
    }

do_auto:
    at_response_free(atResponse);
    atResponse = NULL;

    /* This command does two things, one it sets automatic mode,
       two it starts a new network scan! */
    err = at_send_command("AT+COPS=0", NULL);
    if (err != 0)
        goto error;

finish_scan:

    at_response_free(atResponse);
    atResponse = NULL;

    poll_params->loopcount = 0;
    poll_params->t = t;

    RIL_requestTimedCallback(pollOperatorSelected,
                    poll_params, &TIMEVAL_OPERATOR_SELECT_POLL);

    return;

error:
    free(poll_params);
    at_response_free(atResponse);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    return;    
}

/**
 *  Add handler for RIL_REQUEST_SET_NETWORK_SELECTION_MANUAL
 *  wythe add on 2014-3-28
 *  Set the network selection manually.
 */
void requestSetNetworkSelectionManual(void *data, size_t datalen,
                                      RIL_Token t)
{
    /*
     * AT+COPS=[<mode>[,<format>[,<oper>[,<AcT>]]]]
     *    <mode>   = 4 = Manual (<oper> field shall be present and AcT optionally) with fallback to automatic if manual fails.
     *    <format> = 2 = Numeric <oper>, the number has structure:
     *                   (country code digit 3)(country code digit 2)(country code digit 1)
     *                   (network code digit 2)(network code digit 1)
     */

    (void) datalen;
    int err = 0;
		char *cmd = NULL;
    const char *mccMnc = (const char *) data;

    /* Check inparameter. */
    if (mccMnc == NULL)
        goto error;

    /* Build and send command. */
		asprintf(&cmd, "AT+COPS=1,2,\"%s\"", mccMnc);
    err = at_send_command(cmd, NULL);
    if (err != 0) {
		err = at_send_command("AT+COPS=0",NULL);
		if(err != 0) 
			goto error;
	}

    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    return;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

/**
 *  Add handler for RIL_REQUEST_QUERY_AVAILABLE_NETWORKS
 *  Wythe 2014-3-28 -> start
 *  Query available network to system. It may takes long
 *  time to scan BaseStation around.
 */
static void requestQueryAvailableNetworks(RIL_Token t)
{
    /**
     * AT+COPS=?
     *   +COPS:[(list of supported<stat>,
     *          long alphanumeric <oper>,
                short alphanumeric <oper>,
                numeric <oper>s[,<Act>]s)]
                [,,(list supported <mode>s),(list of supported <format>s)]
     *
     * <stat>   0   Unknown
     *          1   Operator available
     *          2   Operator current
     *          3   Operator forbidden
     */

    /**
     * long alphanumeric
     * short alphanumeric
     * numeric
     * stat
     */
    #define QUERY_AVAILABLE_NW_PARAMS_NUM   4

    const char *statusTable[]=
    {"unknown", "available", "current", "forbidden"};

    ATResponse *atResponse = NULL;
    int err = 0;
    char *line = NULL;
    int n = 0;
    char **responseArray = NULL;
    int i = 0;

    err = at_send_command_singleline("AT+COPS=?", "+COPS", &atResponse);

    if(err != 0 || atResponse->p_intermediates == NULL){
    		LOGE("%s:%s It's error when querying the available network\n", __func__, line);
		    goto error;
		}

    line = atResponse->p_intermediates->line;

    if(line == NULL)
        goto error;

    LOGD("%s:Read COPS response = %s\n", __func__, line);

    err = at_tok_start(&line);
    if(err < 0) goto error;

    err = at_tok_charcounter(line, "(", &n);
    if(err < 0) goto error;

    //Do not need (list supported <mode>s), (list supported <format>s) here
    if(n < 2) n = 0;
    else n -= 2;

		LOGD("%s:Get %d operator information from modem", __func__, n);

    responseArray = (char **)alloca(n * QUERY_AVAILABLE_NW_PARAMS_NUM * sizeof(char *));
		if(NULL == responseArray) goto error;

    for(i = 0; i < n; i++)
    {
        char *s = NULL;
        char *p_line = NULL;
        char *remaining = NULL;
        char *long_alphanumeric = NULL;
        char *short_alphanumeric = NULL;
        char *numeric = NULL;
        int status = 0;
				int act = 0;

        //get the content in (), like ("CHN-UNICOM")
        s = p_line = at_tok_getElementValue(line, "(", ")", &remaining);
        line = remaining;

        if(p_line == NULL)
        {
            LOGE("ERROR:%s NULL pointer while parsing COPS response\n", __func__);
            break;
        }

        //stat <oper>
        err = at_tok_nextint(&p_line, &status);
        if(err < 0) 
        {
error_response:
            free(s);
            goto error;
        }
				
#if 0
				// If the operator's status is unknown or forbidden, don't pass it to RIL java.
				if(0 == status || 3 == status)
				{
						free(s);
						continue;
				}
#endif   
        //long alphanumeric <oper>
        err = at_tok_nextstr(&p_line, &long_alphanumeric);
        if(err < 0) goto error_response;

        //short alphanumeric <oper>
        err = at_tok_nextstr(&p_line, &short_alphanumeric);
        if(err < 0) goto error_response;

        //numeric <oper>
        err = at_tok_nextstr(&p_line, &numeric);
        if(err < 0) goto error_response;

				// Act <oper>
				err = at_tok_nextint(&p_line, &act);
				if(err < 0) goto error_response;

        //add long alphanumeric in array[0]
        responseArray[i*QUERY_AVAILABLE_NW_PARAMS_NUM + 0] = (char *)alloca(strlen(long_alphanumeric)+1);
        strcpy(responseArray[i*QUERY_AVAILABLE_NW_PARAMS_NUM + 0], long_alphanumeric);

        //add short alphanumeric in array[1]
        responseArray[i*QUERY_AVAILABLE_NW_PARAMS_NUM + 1] = (char *)alloca(strlen(short_alphanumeric)+1);
        strcpy(responseArray[i*QUERY_AVAILABLE_NW_PARAMS_NUM + 1], short_alphanumeric);

        //add numeric in array[2]
        responseArray[i*QUERY_AVAILABLE_NW_PARAMS_NUM + 2] = (char *)alloca(strlen(numeric)+1);
        strcpy(responseArray[i*QUERY_AVAILABLE_NW_PARAMS_NUM + 2], numeric);

        free(s);
        
        /**
         *  If module responds empty sring for long alphanumeric and short alphanumeric,
         *  then we copy the MNC/MMC to them.
         */
        
        if (responseArray[i * QUERY_AVAILABLE_NW_PARAMS_NUM + 0] && strlen(responseArray[i * QUERY_AVAILABLE_NW_PARAMS_NUM + 0]) == 0) {
            responseArray[i * QUERY_AVAILABLE_NW_PARAMS_NUM + 0] = (char*) alloca(strlen(responseArray[i * QUERY_AVAILABLE_NW_PARAMS_NUM + 2]) + 1);
            strcpy(responseArray[i * QUERY_AVAILABLE_NW_PARAMS_NUM + 0], responseArray[i * QUERY_AVAILABLE_NW_PARAMS_NUM + 2]);
        }

        if (responseArray[i * QUERY_AVAILABLE_NW_PARAMS_NUM + 1] && strlen(responseArray[i * QUERY_AVAILABLE_NW_PARAMS_NUM + 1]) == 0) {
            responseArray[i * QUERY_AVAILABLE_NW_PARAMS_NUM + 1] = (char*) alloca(strlen(responseArray[i * QUERY_AVAILABLE_NW_PARAMS_NUM + 2]) + 1);
            strcpy(responseArray[i * QUERY_AVAILABLE_NW_PARAMS_NUM + 1], responseArray[i * QUERY_AVAILABLE_NW_PARAMS_NUM + 2]);
        }

				//add status in array[3]
        responseArray[i * QUERY_AVAILABLE_NW_PARAMS_NUM + 3] = (char*) alloca(strlen(statusTable[status])+1);
				strcpy(responseArray[i * QUERY_AVAILABLE_NW_PARAMS_NUM + 3],statusTable[status]);
    }
    
    RIL_onRequestComplete(t, RIL_E_SUCCESS, responseArray, i * QUERY_AVAILABLE_NW_PARAMS_NUM * sizeof(char *));

finally:
    at_response_free(atResponse);
    return;
    
error:
    LOGE("ERROR:requestQueryAvailableNetworks() get no response form module\n");
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	goto finally;
}

/**
 *  Wythe 2014-3-28 -> end
 */

#define MAX_NUM_NEIGHBOR_CELLS 10

void requestGetNeighboringCellIDs(RIL_Token t)
{
#if 0
//{{{
    int network = -1;
    int dummy = 0;
    char *dummyStr = NULL;
    int err = 0;
    ATResponse *atresponse = NULL;
    char *line = NULL;

    /* Determine GSM or WCDMA */
        err = at_send_command_singleline("AT+COPS?", "+COPS:", &atresponse);
    if (err < 0 || atresponse->success == 0)
        goto error;

    line = atresponse->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0)
        goto error;

    err = at_tok_nextint(&line, &dummy);
    if (err < 0)
        goto error;

    err = at_tok_nextint(&line, &dummy);
    if (err < 0)
        goto error;

    err = at_tok_nextstr(&line, &dummyStr);
    if (err < 0)
        goto error;

    err = at_tok_nextint(&line, &network);
    if (err < 0)
        goto error;

    if (/* GSM */ network == 0)
        GSMNeighboringCellIDs(data, datalen, t);
    else if (/* WCDMA */ network == 2)
        WCDMANeighboringCellIDs(data, datalen, t);
    else
        goto error;

finally:
    at_response_free(atresponse);
    return;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
        goto finally;
//}}}
#endif

    int err = 0;
    char *p = NULL;
    int n = 0;
    ATLine *tmp = NULL;
    ATResponse *atresponse = NULL;
    RIL_NeighboringCell *ptr_cells[MAX_NUM_NEIGHBOR_CELLS];

    err = at_send_command_multiline("AT+QENG=\"neighbourcell\"", "+QENG:", &atresponse);
    if (err < 0 || atresponse->success == 0)
        goto error;

    tmp = atresponse->p_intermediates;
    while (tmp) 
    {
        if (n > MAX_NUM_NEIGHBOR_CELLS)
            goto error;
        p = tmp->line;
        {
            char *line = p;
            int lac = 0;
            int cellid = 0;
            int rssi = 0;
            int psc = 0;
            //int pathloss = 0;
            char *net_type;
            char *tmp_str;
            int tmp_int;

            err = at_tok_start(&line);
            if (err < 0)
                goto error;

            /* neighbourcell */
            err = at_tok_nextstr(&line, &tmp_str);
            if (err < 0)
                goto error;

            /* 3G/2G */
            err = at_tok_nextstr(&line, &net_type);
            if (err < 0)
                goto error;

            /* mmc */
            err = at_tok_nextint(&line, &tmp_int);
            if (err < 0)
            {
                LOGE("get mmc error");
                goto next;
            }

            /* mnc */
            err = at_tok_nextint(&line, &tmp_int);
            if (err < 0)
            {
                LOGE("get mnc error");
                goto next;
            }

            /* lac */
            //err = at_tok_nextint(&line, &lac);
            err = at_tok_nexthexint(&line, &lac);
            if (err < 0)
            {
                LOGE("get lac error");
                goto next;
            }

            /* cellid */
            //err = at_tok_nextint(&line, &cellid);
            err = at_tok_nexthexint(&line, &cellid);
            if (err < 0)
            {
                LOGE("get cellid error");
                goto next;
            }

            /* bsic/uarfcn */
            err = at_tok_nextint(&line, &tmp_int);
            if (err < 0)
            {
                LOGE("get bsic/uarfcn error");
                goto next;
            }

            /* psc/arfcn */
            err = at_tok_nextint(&line, &psc);
            if (err < 0)
            {
                LOGE("get psc/arfcn error");
                goto next;
            }

            /* rscp/rxlev */
            err = at_tok_nextint(&line, &rssi);
            if (err < 0)
            {
                LOGE("get rscp/rxlev error");
                goto next;
            }

            /* process data for each cell */
            ptr_cells[n] = alloca(sizeof(RIL_NeighboringCell));
            ptr_cells[n]->rssi = rssi;
            ptr_cells[n]->cid = alloca(9 * sizeof(char));
            if(!strncmp(net_type,"3G",2))
                sprintf(ptr_cells[n]->cid, "%08x", psc);
            else //2G
                sprintf(ptr_cells[n]->cid, "%08x", ((lac << 16) + cellid));
            LOGI("CID:%s  RSSI:%d",ptr_cells[n]->cid,ptr_cells[n]->rssi);
            n++;
        }
next:
        tmp = tmp->p_next;
    }

    RIL_onRequestComplete(t, RIL_E_SUCCESS, ptr_cells,n * sizeof(RIL_NeighboringCell *));

finally:
        at_response_free(atresponse);
    return;

error:
    LOGE("%s error",__func__);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    goto finally;

}

/*** Callback methods from the RIL library to us ***/

/**
 * Call from RIL to us to make a RIL_REQUEST
 *
 * Must be completed with a call to RIL_onRequestComplete()
 *
 * RIL_onRequestComplete() may be called from any thread, before or after
 * this function returns.
 *
 * Will always be called from the same thread, so returning here implies
 * that the radio is ready to process another command (whether or not
 * the previous command has completed).
 */
static void
onRequest (int request, void *data, size_t datalen, RIL_Token t)
{
    ATResponse *p_response;
    int err;

    LOGD("onRequest: %s", requestToString(request));

#if 1 //quectel  //frameworks\base\telephony\java\com\android\internal\telephony/RIL.java ->processUnsolicited()
//case RIL_UNSOL_RIL_CONNECTED : setRadioPower(false, null);
//it it no need to power off radio when RIL.java connect
    onRequestCount++;
    if ((onRequestCount == 1) && (request == RIL_REQUEST_RADIO_POWER) && (((int *)data)[0] == 0)) {
        RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
        return;
    }

    
    if(RADIO_STATE_SIM_LOCKED_OR_ABSENT == currentState()) {
        if (request == RIL_REQUEST_SCREEN_STATE) {
             RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
             return;
        } else if ((request == RIL_REQUEST_DATA_REGISTRATION_STATE) || (request == RIL_REQUEST_VOICE_REGISTRATION_STATE)) {
            char * response[] = {"0"}; //Not registered
            RIL_onRequestComplete(t, RIL_E_SUCCESS, response, sizeof(response)); 
            return;           
        } else if (request == RIL_REQUEST_OPERATOR) {
            char *response[] = {NULL, NULL, NULL};
            RIL_onRequestComplete(t, RIL_E_SUCCESS, response, sizeof(response)); 
            return;           
        } else if (request == RIL_REQUEST_QUERY_NETWORK_SELECTION_MODE) {
            int response[] = {0};
            RIL_onRequestComplete(t, RIL_E_SUCCESS, response, sizeof(response)); 
            return;           
        }
    }

#endif

    /**
     *  Wythe: Add on 2013-04-02 for 4.0 RIL
     *  can not operate SIM Card when sim is not ready.
     *  It's just return GENERIC_FAILURE.
     */
    if(RADIO_STATE_SIM_READY != sState
        && (RIL_REQUEST_WRITE_SMS_TO_SIM == request ||
           RIL_REQUEST_DELETE_SMS_ON_SIM == request)){
        LOGD("radio state =%d, onRequest: %s\r\n", sState, requestToString(request));
        RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
        return;
    }

    /**
     *  Wythe:Add on 2013-04-02 for 4.0 RIL
     *  It's able to get sim status even the radio 
     *  is unavailable.
     */
    if(RADIO_STATE_UNAVAILABLE == sState
        && RIL_REQUEST_GET_SIM_STATUS != request){
        LOGD("radio state =%d, onRequest: %s\r\n", sState, requestToString(request));
        RIL_onRequestComplete(t,RIL_E_RADIO_NOT_AVAILABLE,NULL,0);
        return;
    }

    /**
     *  Wythe:Add on 2013-04-02 for 4.0 RIL
     *  These commands can be executed even the radio is in STATE_OFF
     *  or STATE_SIM_NOT_READY.
     */
    if((RADIO_STATE_OFF == sState || RADIO_STATE_SIM_NOT_READY == sState)
        && !(RIL_REQUEST_RADIO_POWER == request ||
             RIL_REQUEST_GET_SIM_STATUS == request ||
             RIL_REQUEST_STK_GET_PROFILE == request ||
             RIL_REQUEST_STK_SET_PROFILE == request ||
             RIL_REQUEST_REPORT_STK_SERVICE_IS_RUNNING == request ||
             RIL_REQUEST_GET_IMEI == request ||
             RIL_REQUEST_GET_IMEISV ==request ||
             RIL_REQUEST_BASEBAND_VERSION ==request ||
             RIL_REQUEST_SCREEN_STATE == request)){
        LOGD("radio state =%d, onRequest: %s\r\n", sState, requestToString(request));     
        RIL_onRequestComplete(t, RIL_E_RADIO_NOT_AVAILABLE, NULL, 0);
        return;
    }

    /**
     *  Wythe: Add on 2013-04-02 for 4.0 ril
     *  When SIM Card is locked or absent, it's not able to
     *  execute these commands. We just return GENERIC_FAILURE for these requests.
     */
    if(RADIO_STATE_SIM_LOCKED_OR_ABSENT == sState
        && !(RIL_REQUEST_ENTER_SIM_PIN == request ||
             RIL_REQUEST_ENTER_SIM_PIN2 == request ||
             RIL_REQUEST_ENTER_SIM_PUK == request ||
             RIL_REQUEST_ENTER_SIM_PUK2 == request ||
             RIL_REQUEST_ENTER_NETWORK_DEPERSONALIZATION ==request ||
             RIL_REQUEST_GET_SIM_STATUS == request ||
             RIL_REQUEST_RADIO_POWER == request ||
             RIL_REQUEST_GET_IMEISV == request ||
             RIL_REQUEST_GET_IMEI == request ||
             RIL_REQUEST_BASEBAND_VERSION == request ||
             RIL_REQUEST_GET_CURRENT_CALLS == request)){
        LOGD("radio state =%d, onRequest: %s\r\n", sState, requestToString(request));
        RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
        return;
    }

    switch (request) {

        /**** call ****/
        //--------------------------------------------------------
        case RIL_REQUEST_GET_CURRENT_CALLS:
            requestGetCurrentCalls(data, datalen, t);
            break;
            
        case RIL_REQUEST_DIAL:
            requestDial(data, datalen, t);
            break;
            
        case RIL_REQUEST_HANGUP:
            requestHangup(data, datalen, t);
            break;
            
        case RIL_REQUEST_HANGUP_WAITING_OR_BACKGROUND:
            at_send_command("AT+CHLD=0", NULL);

            RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
            break;
            
        case RIL_REQUEST_HANGUP_FOREGROUND_RESUME_BACKGROUND:
            at_send_command("AT+CHLD=1", NULL);

            RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
            break;
            
        case RIL_REQUEST_SWITCH_WAITING_OR_HOLDING_AND_ACTIVE:
            at_send_command("AT+CHLD=2", NULL);

            RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
            break;
            
        case RIL_REQUEST_ANSWER:
            at_send_command("ATA", NULL);

            RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
            break;
            
        case RIL_REQUEST_CONFERENCE:
            at_send_command("AT+CHLD=3", NULL);

            RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
            break;
            
        case RIL_REQUEST_UDUB:
            at_send_command("ATH", NULL);

            RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
            break;

        case RIL_REQUEST_SEPARATE_CONNECTION:
            {
                char  cmd[12];
                int   party = ((int*)data)[0];

                if (party > 0 && party < 10) {
                    sprintf(cmd, "AT+CHLD=2%d", party);
                    at_send_command(cmd, NULL);
                    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
                } else {
                    LOGD("%s error for RIL_REQUEST_SEPARATE_CONNECTION!\n", __func__);
                    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
                }
            }
            break;
        //--------------------------------------------------------    

        /**** sms ****/
        //--------------------------------------------------------
        case RIL_REQUEST_SEND_SMS:
            requestSendSMS(data, datalen, t);
            break;

        case RIL_REQUEST_SEND_SMS_EXPECT_MORE:
            requestSendSMSExpectMore(data, datalen, t);
            break;

        case RIL_REQUEST_SMS_ACKNOWLEDGE:
            requestSMSAcknowledge(data, datalen, t);
            break;

        case RIL_REQUEST_WRITE_SMS_TO_SIM:
            requestWriteSmsToSim(data, datalen, t);
            break;

        case RIL_REQUEST_DELETE_SMS_ON_SIM: {
            char * cmd;
            p_response = NULL;
            asprintf(&cmd, "AT+CMGD=%d", ((int *)data)[0] - 1);
            err = at_send_command(cmd, &p_response);
            free(cmd);
            if (err < 0 || p_response->success == 0) {
                LOGE("%s error for RIL_REQUEST_DELETE_SMS_ON_SIM!\n", __func__);
                RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
            } else {
                RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
            }
            at_response_free(p_response);
            break;
        }
        //--------------------------------------------------------

        /**** data connection ****/
        //--------------------------------------------------------
        case RIL_REQUEST_DEACTIVATE_DATA_CALL: {
            requestDeactivateDefaultPDP(data, datalen, t);
            break;
        }

        case RIL_REQUEST_LAST_DATA_CALL_FAIL_CAUSE: {
            requestLastPDPFailCause(t);
            break;
        }

        case RIL_REQUEST_SETUP_DATA_CALL:
            requestSetupDataCall(data, datalen, t);
            break;

        case RIL_REQUEST_DATA_CALL_LIST:
            requestDataCallList(data, datalen, t);
            break;
        //--------------------------------------------------------    

        /**** network ****/
        //--------------------------------------------------------
        case RIL_REQUEST_SET_PREFERRED_NETWORK_TYPE: {
        if (ql_is_UC20)
            requestSetPreferredNetworkType(data,datalen,t);
        else
            	RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);
            break;
        }

        case RIL_REQUEST_SIGNAL_STRENGTH:
            requestSignalStrength(data, datalen, t);
            break;
            
        case RIL_REQUEST_VOICE_REGISTRATION_STATE:
        case RIL_REQUEST_DATA_REGISTRATION_STATE:
            requestRegistrationState(request, data, datalen, t);
            break;
            
        case RIL_REQUEST_OPERATOR:
            requestOperator(data, datalen, t);
            break;

				/***
				 *	Drop.
				 *	wythe 2014-3-28 
				 */
#if 0
        case RIL_REQUEST_SET_NETWORK_SELECTION_AUTOMATIC:
            at_send_command("AT+COPS=0", NULL);
            break;
#endif

        case RIL_REQUEST_QUERY_NETWORK_SELECTION_MODE:
            requestQueryNetworkSelectionMode(data, datalen, t);
            break;

        /**
         *  Add new request:RIL_REQUEST_GET_PREFERRED_NETWORK_TYPE
         *  Wythe 2013-9-27
         */
        case RIL_REQUEST_GET_PREFERRED_NETWORK_TYPE:
        if (ql_is_UC20)
            requestGetPreferredNetworkType(t);
        else
            	RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);
         break;

        /**
         *  Add new request:RIL_REQUEST_QUERY_AVAILABLE_NETWORKS
         *  Wythe 2014-3-28
         */
        case RIL_REQUEST_QUERY_AVAILABLE_NETWORKS:
            requestQueryAvailableNetworks(t);
            break;

        /**
         *  Add new request:RIL_REQUEST_SET_NETWORK_SELECTION_MANUAL
         *  Wythe 2014-3-28
         */
        case RIL_REQUEST_SET_NETWORK_SELECTION_MANUAL:
            requestSetNetworkSelectionManual(data, datalen, t);
            break;

        /**
         *  Add new request:RIL_REQUEST_SET_NETWORK_SELECTION_AUTOMATIC
         *  Wythe 2014-3-28
         *
         */
        case RIL_REQUEST_SET_NETWORK_SELECTION_AUTOMATIC:
            requestSetNetworkSelectionAutomatic(t);
            break;
            
        //--------------------------------------------------------    

        /**** sim card ****/
        //--------------------------------------------------------
        case RIL_REQUEST_GET_SIM_STATUS: {
            RIL_CardStatus_v6 *p_card_status;
            char *p_buffer;
            int buffer_size;

            int result = getCardStatus(&p_card_status);
            if (result == RIL_E_SUCCESS) {
                p_buffer = (char *)p_card_status;
                buffer_size = sizeof(*p_card_status);
            } else {
                p_buffer = NULL;
                buffer_size = 0;
            }
            RIL_onRequestComplete(t, result, p_buffer, buffer_size);
            freeCardStatus(p_card_status);
            break;
        }

         case RIL_REQUEST_GET_IMSI:
            p_response = NULL;
            err = at_send_command_numeric("AT+CIMI", &p_response);

            if (err < 0 || p_response->success == 0) {
                LOGE("%s error for RIL_REQUEST_GET_IMSI!\n", __func__);
                RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
            } else {
                RIL_onRequestComplete(t, RIL_E_SUCCESS,
                    p_response->p_intermediates->line, sizeof(char *));
            }
            at_response_free(p_response);
            break;

         case RIL_REQUEST_SIM_IO:
            requestSIM_IO(data,datalen,t);
            break;

        case RIL_REQUEST_QUERY_FACILITY_LOCK:
            requestQuestFacility(data,datalen,t);
            break;

        case RIL_REQUEST_SET_FACILITY_LOCK:
            requestSetFacility(data,datalen,t,request);
            break;

        case RIL_REQUEST_CHANGE_SIM_PIN:
            requestChangeSimPin(data,datalen,"SC",t,request);
            break;

        case RIL_REQUEST_CHANGE_SIM_PIN2:
            requestChangeSimPin(data,datalen,"P2",t,request);
            break;

        case RIL_REQUEST_ENTER_SIM_PIN:
        case RIL_REQUEST_ENTER_SIM_PUK:
        case RIL_REQUEST_ENTER_SIM_PIN2:
        case RIL_REQUEST_ENTER_SIM_PUK2:
            requestEnterSimPin(data, datalen, t,request);
            break;
        //--------------------------------------------------------    
            
        /**** other,e.g. system, baseband information****/
        case RIL_REQUEST_SCREEN_STATE: {
            requestScreenState(data,datalen,t);
            break;
        }

        case RIL_REQUEST_BASEBAND_VERSION: {
            requestBaseBandVersion(data,datalen,t);
            break;
        }

        case RIL_REQUEST_GET_IMEISV: {
            requestGetIMEISV(t);
            break;
        }

        /**
         *  Not support STK now.
         *  Wythe 2013-9-26
         */
#if 0         
        case RIL_REQUEST_REPORT_STK_SERVICE_IS_RUNNING: {
            requestReportSTKServiceIsRunning(t);
            break;
        }
#endif

        case RIL_REQUEST_RADIO_POWER:
            requestRadioPower(data, datalen, t);
            break;
        case RIL_REQUEST_DTMF: {
            char c = ((char *)data)[0];
            char *cmd;
            asprintf(&cmd, "AT+VTS=%c", (int)c);
            at_send_command(cmd, NULL);
            free(cmd);
            RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
            break;
        }

        case RIL_REQUEST_GET_IMEI:
            if (ql_is_EC20) { //EC20 bug, canot get OK
				RIL_onRequestComplete(t, RIL_E_SUCCESS, "355189036244202", sizeof(char *));
				return;
			}
            p_response = NULL;
            err = at_send_command_numeric("AT+CGSN", &p_response);

            if (err < 0 || p_response->success == 0) {
                LOGE("%s error for RIL_REQUEST_GET_IMEI!\n", __func__);
                RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
            } else {
                RIL_onRequestComplete(t, RIL_E_SUCCESS,
                    p_response->p_intermediates->line, sizeof(char *));
            }
            at_response_free(p_response);
            break;

        case RIL_REQUEST_SEND_USSD:
            requestSendUSSD(data, datalen, t);
            break;

        case RIL_REQUEST_CANCEL_USSD:
#if 1 //quectel
            ussd_pending_index++;
#endif
            at_send_command("AT+CUSD=2", NULL);
            RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
            break;

        case RIL_REQUEST_OEM_HOOK_RAW:
            // echo back data
            RIL_onRequestComplete(t, RIL_E_SUCCESS, data, datalen);
            break;

        case RIL_REQUEST_SET_BAND_MODE:
            if (ql_is_UC20) {
                ATResponse *atResponse = NULL;
                int err;
                int bandval = 0;
                char *cmd;
                
                switch(((int *)data)[0])
                {
                    case 1:
                        bandval = 3;
                        break;
                    case 2:
                        bandval = 108;
                        break;
                    case 3:
                        bandval = 256;
                        break;
                    case 4:
                    case 5:
                        bandval = 67;
                        break;
                    default:
                        bandval = 512;
                }
                asprintf(&cmd,"AT+QCFG=\"band\",%d",bandval);
                err = at_send_command(cmd,&atResponse);

                if(err < 0 || atResponse == NULL || atResponse->success == 0)
                    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
                else
                    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
                free(atResponse);
                free(cmd);
            } else {
                RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);
            }
            break;

        case RIL_REQUEST_QUERY_AVAILABLE_BAND_MODE:
            RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);
            break;

        case RIL_REQUEST_SET_TTY_MODE:
            RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);
            break;

        case RIL_REQUEST_GET_SMSC_ADDRESS:
            {
                ATResponse *p_response = NULL;
                char *cmd = NULL; 
                char *SMSC_address = NULL;
                int err;

                err = at_send_command_singleline("AT+CSCA?","+CSCA:",&p_response);
                if(err < 0 || p_response == NULL || p_response->success == 0)
                    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
                else
                {
                    char *line;

                    line = p_response->p_intermediates->line;
                    err = at_tok_start(&line);
                    err = at_tok_nextstr(&line,&SMSC_address);
                    RIL_onRequestComplete(t, RIL_E_SUCCESS, SMSC_address, sizeof(char *));
                }
                free(p_response);
                free(cmd);
            }
            break;

        case RIL_REQUEST_SET_SMSC_ADDRESS:
            {
                ATResponse *p_response = NULL;
                char *cmd = NULL; 
                char *SMSC_address = NULL;
                int err;
                
                SMSC_address = data;
                asprintf(&cmd,"AT+CSCA=\"%s\"",SMSC_address);

                err = at_send_command(cmd,&p_response);
                if(err < 0 || p_response == NULL || p_response->success == 0)
                    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
                else
                    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);

                free(p_response);
                free(cmd);
            }
            break;


        case RIL_REQUEST_OEM_HOOK_STRINGS: {
            int i;
            const char ** cur;

            LOGD("got OEM_HOOK_STRINGS: 0x%8p %lu", data, (long)datalen);


            for (i = (datalen / sizeof (char *)), cur = (const char **)data ;
                    i > 0 ; cur++, i --) {
                LOGD("> '%s'", *cur);
            }

#if 1 //quectel
{
        int err;
        ATResponse *p_response = NULL;
        err = at_send_command_multiline( ((const char **)data)[0], "\0", &p_response);
        if(err < 0 || p_response == NULL || p_response->success == 0)
            RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
        else
        {
            char *lines[100] ;
            size_t i = 0;
            ATLine  *p_intermediates = p_response->p_intermediates;
            while (p_intermediates && (i < ((sizeof(lines) / sizeof(lines[0])) - 1))) {
                if (p_intermediates->line)
                    lines[i++] = p_intermediates->line;
                p_intermediates = p_intermediates->p_next;
            }
            if (p_response->finalResponse)
                lines[i++] = p_response->finalResponse;
            RIL_onRequestComplete(t, RIL_E_SUCCESS, lines, i*sizeof(char *));
        }
}
#else
            // echo back strings
            RIL_onRequestComplete(t, RIL_E_SUCCESS, data, datalen);
#endif
            break;
        }

        case RIL_REQUEST_GSM_SET_BROADCAST_SMS_CONFIG: {
            requestGsmSetBroadcastSmsConfig(data, datalen, t);
            break;
        }

        case RIL_REQUEST_GSM_GET_BROADCAST_SMS_CONFIG: {
            requestGsmGetBroadcastSmsConfig(t);
            break;
        }

        case RIL_REQUEST_GSM_SMS_BROADCAST_ACTIVATION: {
            requestGsmSmsBroadcastActivation(data, datalen, t);
            break;
        }

        case RIL_REQUEST_RESET_RADIO: {
        if (ql_is_UC20)            
            requestResetRadio(t);
        else
            	RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);
            break;
        }

        case RIL_REQUEST_GET_MUTE: {
            requestGetMute(t);
            break;
        }

        case RIL_REQUEST_SET_MUTE: {
            requestSetMute(data, datalen, t);
            break;
        }

        case RIL_REQUEST_DTMF_START: {
            requestDtmfStart(data, datalen, t);
            break;
        }

        case RIL_REQUEST_DTMF_STOP: {
            requestDtmfStop(t);
            break;
        }

        case RIL_REQUEST_QUERY_CALL_FORWARD_STATUS: {
            requestQueryCallForwardStatus(data, datalen, t);
            break;
        }

        case RIL_REQUEST_SET_CALL_FORWARD: {
            requestSetCallForward(data, datalen, t);
            break;
        }

        case RIL_REQUEST_GET_CLIR: {
            requestGetClir(t);
            break;
        }

        case RIL_REQUEST_SET_CLIR: {
            requestSetClir(data, datalen, t);
            break;
        }

        case RIL_REQUEST_QUERY_CALL_WAITING: {
            requestQueryCallWaiting(data, datalen, t);
            break;
        }

        case RIL_REQUEST_SET_CALL_WAITING: {
            requestSetCallWaiting(data, datalen, t);
            break;
        }
#if 0
        case RIL_REQUEST_QUERY_AVAILABLE_NETWORKS:
            requestQueryAvailableNetworks(t);
            break;

        case RIL_REQUEST_SET_NETWORK_SELECTION_MANUAL:
            requestSetNetworkSelectionManual(data, datalen, t);
            break;

        case RIL_REQUEST_SET_NETWORK_SELECTION_AUTOMATIC:
            requestSetNetworkSelectionAutomatic(t);
            break;
#endif
        case RIL_REQUEST_GET_NEIGHBORING_CELL_IDS:
        if (ql_is_UC20)
            requestGetNeighboringCellIDs(t);
        else
              RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);          
            break;

#ifdef RIL_REQUEST_QUECTEL_AT
        case RIL_REQUEST_QUECTEL_AT: {
		ATResponse *p_response = NULL;
		int err = at_send_command((const char *)data,&p_response);
		//if(err < 0 || p_response == NULL || p_response->success == 0)
		//RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
		//else
		//RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
		free(p_response);
		break;
	}
#endif

        default:
            RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);
            break;
    }
}

/**
 * Synchronous call from the RIL to us to return current radio state.
 * RADIO_STATE_UNAVAILABLE should be the initial state.
 */
static RIL_RadioState
currentState()
{
    return sState;
}
/**
 * Call from RIL to us to find out whether a specific request code
 * is supported by this implementation.
 *
 * Return 1 for "supported" and 0 for "unsupported"
 */

static int
onSupports (int requestCode)
{
    (void ) requestCode;
    LOGI("Call OnSupports\r\n");
    return 1;
}

static void onCancel (RIL_Token t)
{
    (void) t;
    LOGI("Call OnCancel\r\n");
}

static const char * getVersion(void)
{
#if 1 //quectel
    onRequestCount = 0; //onNewCommandConnect will call this function, and RIL.java will send RIL_REQUEST_RADIO_POWER
#endif
    return REFERENCE_RIL_VERSION;
}

static void
setRadioState(RIL_RadioState newState)
{
    RIL_RadioState oldState;

    pthread_mutex_lock(&s_state_mutex);

    oldState = sState;

    LOGI("[%s]:oldState=%d, newState=%d\r\n", __func__, oldState, newState);

    if (s_closed > 0) {
        // If we're closed, the only reasonable state is
        // RADIO_STATE_UNAVAILABLE
        // This is here because things on the main thread
        // may attempt to change the radio state after the closed
        // event happened in another thread
        newState = RADIO_STATE_UNAVAILABLE;
    }

    if (sState != newState || s_closed > 0) {
        sState = newState;

        pthread_cond_broadcast (&s_state_cond);
    }

    pthread_mutex_unlock(&s_state_mutex);


    /* do these outside of the mutex */
    if (sState != oldState) {
        RIL_onUnsolicitedResponse (RIL_UNSOL_RESPONSE_RADIO_STATE_CHANGED,
                                    NULL, 0);

        /* FIXME onSimReady() and onRadioPowerOn() cannot be called
         * from the AT reader thread
         * Currently, this doesn't happen, but if that changes then these
         * will need to be dispatched on the request thread
         */
        if (sState == RADIO_STATE_SIM_READY) {
            onSIMReady();
        } else if (sState == RADIO_STATE_SIM_NOT_READY) {
            onRadioPowerOn();
        }
    }
}

/** Returns SIM_NOT_READY on error */
static SIM_Status
getSIMStatus()
{
    ATResponse *p_response = NULL;
    int err;
    int ret;
    char *cpinLine;
    char *cpinResult;

    if (sState == RADIO_STATE_OFF || sState == RADIO_STATE_UNAVAILABLE) {
        ret = SIM_NOT_READY;
        goto done;
    }

    err = at_send_command_singleline("AT+CPIN?", "+CPIN:", &p_response);

    if (err != 0) {
        ret = SIM_NOT_READY;
        goto done;
    }

    switch (at_get_cme_error(p_response)) {
        case CME_SUCCESS:
            break;

        case CME_SIM_NOT_INSERTED:
            ret = SIM_ABSENT;
            goto done;

        default:
            ret = SIM_NOT_READY;
            goto done;
    }

    /* CPIN? has succeeded, now look at the result */

    cpinLine = p_response->p_intermediates->line;
    err = at_tok_start (&cpinLine);

    if (err < 0) {
        ret = SIM_NOT_READY;
        goto done;
    }

    err = at_tok_nextstr(&cpinLine, &cpinResult);

    if (err < 0) {
        ret = SIM_NOT_READY;
        goto done;
    }

    if (0 == strcmp (cpinResult, "SIM PIN")) {
        ret = SIM_PIN;
        goto done;
    } else if (0 == strcmp (cpinResult, "SIM PUK")) {
        ret = SIM_PUK;
        goto done;
    } else if (0 == strcmp (cpinResult, "PH-NET PIN")) {
        return SIM_NETWORK_PERSONALIZATION;
    } else if (0 != strcmp (cpinResult, "READY"))  {
        /* we're treating unsupported lock types as "sim absent" */
        ret = SIM_ABSENT;
        goto done;
    }

    at_response_free(p_response);
    p_response = NULL;
    cpinResult = NULL;

    ret = SIM_READY;

done:
    at_response_free(p_response);
    return ret;
}


/**
 * Get the current card status.
 *
 * This must be freed using freeCardStatus.
 * @return: On success returns RIL_E_SUCCESS
 */
static int getCardStatus(RIL_CardStatus_v6 **pp_card_status) {
    static RIL_AppStatus app_status_array[] = {
        // SIM_ABSENT = 0
        { RIL_APPTYPE_UNKNOWN, RIL_APPSTATE_UNKNOWN, RIL_PERSOSUBSTATE_UNKNOWN,
          NULL, NULL, 0, RIL_PINSTATE_UNKNOWN, RIL_PINSTATE_UNKNOWN },
        // SIM_NOT_READY = 1
        { RIL_APPTYPE_SIM, RIL_APPSTATE_DETECTED, RIL_PERSOSUBSTATE_UNKNOWN,
          NULL, NULL, 0, RIL_PINSTATE_UNKNOWN, RIL_PINSTATE_UNKNOWN },
        // SIM_READY = 2
        { RIL_APPTYPE_SIM, RIL_APPSTATE_READY, RIL_PERSOSUBSTATE_READY,
          NULL, NULL, 0, RIL_PINSTATE_UNKNOWN, RIL_PINSTATE_UNKNOWN },
        // SIM_PIN = 3
        { RIL_APPTYPE_SIM, RIL_APPSTATE_PIN, RIL_PERSOSUBSTATE_UNKNOWN,
          NULL, NULL, 0, RIL_PINSTATE_ENABLED_NOT_VERIFIED, RIL_PINSTATE_UNKNOWN },
        // SIM_PUK = 4
        { RIL_APPTYPE_SIM, RIL_APPSTATE_PUK, RIL_PERSOSUBSTATE_UNKNOWN,
          NULL, NULL, 0, RIL_PINSTATE_ENABLED_BLOCKED, RIL_PINSTATE_UNKNOWN },
        // SIM_NETWORK_PERSONALIZATION = 5
        { RIL_APPTYPE_SIM, RIL_APPSTATE_SUBSCRIPTION_PERSO, RIL_PERSOSUBSTATE_SIM_NETWORK,
          NULL, NULL, 0, RIL_PINSTATE_ENABLED_NOT_VERIFIED, RIL_PINSTATE_UNKNOWN }
    };
    RIL_CardState card_state;
    int num_apps;

    int sim_status = getSIMStatus();
    if (sim_status == SIM_ABSENT) {
        card_state = RIL_CARDSTATE_ABSENT;
        num_apps = 0;
    } else {
        card_state = RIL_CARDSTATE_PRESENT;
        num_apps = 1;
    }

    // Allocate and initialize base card status.
    RIL_CardStatus_v6 *p_card_status = malloc(sizeof(RIL_CardStatus_v6));
    p_card_status->card_state = card_state;
    p_card_status->universal_pin_state = RIL_PINSTATE_UNKNOWN;
    p_card_status->gsm_umts_subscription_app_index = RIL_CARD_MAX_APPS;
    p_card_status->cdma_subscription_app_index = RIL_CARD_MAX_APPS;
    p_card_status->ims_subscription_app_index = RIL_CARD_MAX_APPS;
    p_card_status->num_applications = num_apps;

    // Initialize application status
    int i;
    for (i = 0; i < RIL_CARD_MAX_APPS; i++) {
        p_card_status->applications[i] = app_status_array[SIM_ABSENT];
    }

    // Pickup the appropriate application status
    // that reflects sim_status for gsm.
    if (num_apps != 0) {
        // Only support one app, gsm
        p_card_status->num_applications = 1;
        p_card_status->gsm_umts_subscription_app_index = 0;

        // Get the correct app status
        p_card_status->applications[0] = app_status_array[sim_status];
    }

    *pp_card_status = p_card_status;
    return RIL_E_SUCCESS;
}

/**
 * Free the card status returned by getCardStatus
 */
static void freeCardStatus(RIL_CardStatus_v6 *p_card_status) {
    free(p_card_status);
}

/**
 * SIM ready means any commands that access the SIM will work, including:
 *  AT+CPIN, AT+CSMS, AT+CNMI, AT+CRSM
 *  (all SMS-related commands)
 */

static void pollSIMState (void *param)
{
    ATResponse *p_response;
    int ret;

    if (sState != RADIO_STATE_SIM_NOT_READY) {
        // no longer valid to poll
        return;
    }

    switch(getSIMStatus()) {
        case SIM_ABSENT:
        case SIM_PIN:
        case SIM_PUK:
        case SIM_NETWORK_PERSONALIZATION:
        default:
            setRadioState(RADIO_STATE_SIM_LOCKED_OR_ABSENT);
        return;

        case SIM_NOT_READY:
            RIL_requestTimedCallback (pollSIMState, NULL, &TIMEVAL_SIMPOLL);
        return;

        case SIM_READY:
            setRadioState(RADIO_STATE_SIM_READY);
        return;
    }
}

/** returns 1 if on, 0 if off, and -1 on error */
static int isRadioOn()
{
    ATResponse *p_response = NULL;
    int err;
    char *line;
    char ret;

    err = at_send_command_singleline("AT+CFUN?", "+CFUN:", &p_response);

    if (err < 0 || p_response->success == 0) {
        // assume radio is off
        goto error;
    }

    line = p_response->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0) goto error;

    err = at_tok_nextbool(&line, &ret);
    if (err < 0) goto error;

    at_response_free(p_response);

    return (int)ret;

error:
    LOGE("[%s]:isRadioOn return error\n", __func__);
    at_response_free(p_response);
    return -1;
}

/**
 * Initialize everything that can be configured while we're still in
 * AT+CFUN=0
 */
static void initializeCallback(void *param)
{
    ATResponse *p_response = NULL;
    int err;

    setRadioState (RADIO_STATE_OFF);

    at_handshake();

#if 1 //quectel
    sleep(3); //sleep 3 seconnds to wait module initialize
#endif

#if 1 //quectel display module software version 
    ql_is_UC20 = ql_is_UG95 = ql_is_EC20 = 0;
    err = at_send_command_multiline("ATI", "\0", &p_response);
    if (!err && p_response && p_response->success) {
        ATLine *p_cur = p_response->p_intermediates;
        if (p_cur && (p_cur = p_cur->p_next)) {
             if (!strcmp(p_cur->line, "UC15")) {
                ql_is_UC20 = 1;
                LOGD("Quectel Product: UC15");
            } else if (!strcmp(p_cur->line, "UC20")) {
                 ql_is_UC20 = 1;
                LOGD("Quectel Product: UC20");
            } else if (!strcmp(p_cur->line, "UG95")) {
                 ql_is_UG95 = 1;
                LOGD("Quectel Product: UG95");
            } else if (!strcmp(p_cur->line, "UG96")) {
                 ql_is_UG95 = 1;
                LOGD("Quectel Product: UG96");
            } else if (!strcmp(p_cur->line, "EC20")) {
                ql_is_EC20 = 1;
                LOGD("Quectel Product: EC20");
            }
        }
    }
    at_response_free(p_response);
    p_response = NULL;
    if (!(ql_is_UC20|ql_is_UG95|ql_is_EC20))
        LOGD("Quectel Product: unkonw!");
#endif

    /* note: we don't check errors here. Everything important will
       be handled in onATTimeout and onATReaderClosed */

    /*  atchannel is tolerant of echo but it must */
    /*  have verbose result codes */
    at_send_command("ATE0Q0V1", NULL);

    /*  No auto-answer */
    at_send_command("ATS0=0", NULL);

//Joe
#ifdef URC_PORT_UART
    at_send_command("AT+QURCCFG=\"URCPORT\",\"UART1\"", NULL);
#endif

#if 1 // quectel Set DTR Function Mode 
//ON->OFF on DTR: Disconnect data call, change to command mode. During state DTR = OFF, auto-answer function is disabled
    at_send_command("AT&D2", NULL);
#endif

    /*  Extended errors */
    at_send_command("AT+CMEE=1", NULL);

    /*  Network registration events */
    err = at_send_command("AT+CREG=2", &p_response);

    /* some handsets -- in tethered mode -- don't support CREG=2 */
    if (err < 0 || p_response->success == 0) {
        at_send_command("AT+CREG=1", NULL);
    }

    at_response_free(p_response);

#if 1 //quectel
//there is a change in the ME network registration status or a change of the network cell: 
//+CGREG: <stat>[, <lac>,<ci> ] 
    at_send_command("AT+CGREG=2", NULL); 
#else
    /*  GPRS registration events */
    at_send_command("AT+CGREG=1", NULL);
#endif

    /*  Call Waiting notifications */
    at_send_command("AT+CCWA=1", NULL);

    /*  Alternating voice/data off */
    at_send_command("AT+CMOD=0", NULL);

    /*  Not muted */
    at_send_command("AT+CMUT=0", NULL);

    /*  +CSSU unsolicited supp service notifications */
    at_send_command("AT+CSSN=0,1", NULL);

    /*  no connected line identification */
    at_send_command("AT+COLP=0", NULL);

#if 1 //quectel
#ifdef CUSD_USE_UCS2_MODE
    at_send_command("AT+CSCS=\"UCS2\"", NULL);
    if (ql_is_UC20)
        at_send_command("at+qcfg=\"ussd/cause\",1", NULL);
#endif
#else
    at_send_command("AT+CSCS=\"HEX\"", NULL);
#endif

    /*  USSD unsolicited */
    at_send_command("AT+CUSD=1", NULL);

#if 1 //quectel
    at_send_command("AT+CGEREP=0", NULL); //carl tmp disable
#else
    /*  Enable +CGEV GPRS event notifications, but don't buffer */
    at_send_command("AT+CGEREP=1,0", NULL);
#endif

    /*  SMS PDU mode */
    at_send_command("AT+CMGF=0", NULL);

#if 1 //quectel
    if (ql_is_UC20) {
        at_send_command("AT+QINDCFG=\"all\",1", NULL);
        at_send_command("AT+QINDCFG=\"smsfull\",1", NULL);
    }
    //Quectel NITZ
    //Joe.Wang
    //2013-11-14
    at_send_command("AT+CTZU=1", NULL);
    at_send_command("AT+CTZR=2", NULL);

    if (ql_is_UC20 || ql_is_EC20)
        at_send_command("AT+QCFG=\"QMISYNC\",0", NULL);
    at_send_command("AT+QSCLK=1", NULL); //Configure Whether or not to Enter into Sleep Mode
#endif
    
    /* assume radio is off on error */
    if (isRadioOn() > 0) {
        setRadioState (RADIO_STATE_SIM_NOT_READY);
    }
#if 1 //quectel
    else {
        at_send_command("AT+CFUN=1", NULL);
        sleep(5);
    } 
#endif   
}

static void waitForClose()
{
    pthread_mutex_lock(&s_state_mutex);

    while (s_closed == 0) {
        pthread_cond_wait(&s_state_cond, &s_state_mutex);
    }
    
    pthread_mutex_unlock(&s_state_mutex);
}

/**
 * Called by atchannel when an unsolicited line appears
 * This is called on atchannel's reader thread. AT commands may
 * not be issued here
 */
static void onUnsolicited (const char *s, const char *sms_pdu)
{
    char *line = NULL;
    int err;

    /* Ignore unsolicited responses until we're initialized.
     * This is OK because the RIL library will poll for initial state
     */
    if (sState == RADIO_STATE_UNAVAILABLE) {
        return;
    }

    if (strStartsWith(s, "%CTZV:")) {
        /* TI specific -- NITZ time */
        char *response;

        line = strdup(s);
        at_tok_start(&line);

        err = at_tok_nextstr(&line, &response);

        if (err != 0) {
            LOGE("invalid NITZ line %s\n", s);
        } else {
            RIL_onUnsolicitedResponse (
                RIL_UNSOL_NITZ_TIME_RECEIVED,
                response, strlen(response));
        }
    } else if (strStartsWith(s,"+CRING:")
                || strStartsWith(s,"RING")
                || strStartsWith(s,"NO CARRIER")
                || strStartsWith(s,"+CCWA")
    ) {
        RIL_onUnsolicitedResponse (
            RIL_UNSOL_RESPONSE_CALL_STATE_CHANGED,
            NULL, 0);
#ifdef WORKAROUND_FAKE_CGEV
        RIL_requestTimedCallback (onDataCallListChanged, NULL, NULL); //TODO use new function
#endif /* WORKAROUND_FAKE_CGEV */
    } else if (strStartsWith(s,"+CREG:")
                || strStartsWith(s,"+CGREG:")
    ) {
#if 1 //quectel
    if (gprsState && (network_debounce_time == 0))
        network_debounce_time = NETWORK_DEBOUNCE_TIMEOUT;
#endif
        RIL_onUnsolicitedResponse (
            RIL_UNSOL_RESPONSE_VOICE_NETWORK_STATE_CHANGED,
            NULL, 0);
#ifdef WORKAROUND_FAKE_CGEV
        RIL_requestTimedCallback (onDataCallListChanged, NULL, NULL);
#endif /* WORKAROUND_FAKE_CGEV */
    } else if (strStartsWith(s, "+CMT:")) {
        RIL_onUnsolicitedResponse (
            RIL_UNSOL_RESPONSE_NEW_SMS,
            sms_pdu, strlen(sms_pdu));
    } else if (strStartsWith(s, "+CDS:")) {
        RIL_onUnsolicitedResponse (
            RIL_UNSOL_RESPONSE_NEW_SMS_STATUS_REPORT,
            sms_pdu, strlen(sms_pdu));
    } else if (strStartsWith(s, "+CGEV:")) {
        /* Really, we can ignore NW CLASS and ME CLASS events here,
         * but right now we don't since extranous
         * RIL_UNSOL_DATA_CALL_LIST_CHANGED calls are tolerated
         */
        /* can't issue AT commands here -- call on main thread */
        RIL_requestTimedCallback (onDataCallListChanged, NULL, NULL);
#ifdef WORKAROUND_FAKE_CGEV
    } else if (strStartsWith(s, "+CME ERROR: 150")) {
        RIL_requestTimedCallback (onDataCallListChanged, NULL, NULL);
#endif /* WORKAROUND_FAKE_CGEV */
    } 
#if 1 //quectel
    else if(strStartsWith(s,"+CTZE:")) { //NITZ
        //+CTZE: "+32",0,"2014/05/05,05:05:30"
        //yy/mm/dd,hh:mm:ss(+/-)tz,dt
        char *tz = NULL;
        int dt;
        char *n_time = NULL;
        char *response = NULL;

        line = strdup(s);
        at_tok_start(&line);
        err = at_tok_nextstr(&line,&tz);
        if (err < 0) return;
        err = at_tok_nextint(&line,&dt);
        if (err < 0) return;
        err = at_tok_nextstr(&line,&n_time);
        if (err < 0) return;

        if (n_time[4] == '/')
            asprintf(&response,"%s%s,%d",n_time+2,tz,dt); //2014 -> 14
        else    
            asprintf(&response,"%s%s,%d",n_time,tz,dt);
        time_zone_report++;

        RIL_onUnsolicitedResponse (RIL_UNSOL_NITZ_TIME_RECEIVED,response, strlen(response));
    } else if (strStartsWith(s, "+CMTI:")) {
        char *memory = NULL;
        int index;
        line = strdup(s);
        at_tok_start(&line);
        err = at_tok_nextstr(&line,&memory);
        if (err < 0) return;
        if(!strcmp("SM",memory))
        {
            err = at_tok_nextint(&line,&index);
            if (err < 0) return;
             RIL_onUnsolicitedResponse (RIL_UNSOL_RESPONSE_NEW_SMS_ON_SIM,&index, sizeof(int *));
        }
    } else if (strStartsWith(s, "+QIND: \"smsfull\"")) {
        char *memory = NULL;
        char *tmp = NULL;
        line = strdup(s);
        at_tok_start(&line);
        err = at_tok_nextstr(&line,&tmp);
        if (err < 0) return;
        err = at_tok_nextstr(&line,&memory);
        if (err < 0) return;
        if(!strcmp("SM",memory))
        {
             RIL_onUnsolicitedResponse (RIL_UNSOL_SIM_SMS_STORAGE_FULL,NULL,0);
        }
#if 1 // RIL will get sim state by pollSIMState() 
    } else if (strStartsWith(s, "+CPIN: READY") || strStartsWith(s, "+QUSIM:") || strStartsWith(s, "+QIND: SMS DONE") || strStartsWith(s, "+QIND: PB DONE")) {
        if(currentState() == RADIO_STATE_SIM_LOCKED_OR_ABSENT)
            setRadioState(RADIO_STATE_SIM_READY);
#endif
    } else if (strStartsWith(s, "+CPIN: NOT READY")) {
        if(currentState() == RADIO_STATE_SIM_READY)
            setRadioState(RADIO_STATE_SIM_LOCKED_OR_ABSENT);
    } else if (strStartsWith(s, "+CUSD")) {
        char *response[2];
        char *data;
        char *mode;
        int dcs;
        int outlen = 0;
        int err;

        ussd_pending_index++;
        line = strdup(s);
        at_tok_start(&line);
        err = at_tok_nextstr(&line,&mode);
        if(err != 0)
        {
            LOGE("at_tok_nextstr[mode] error");
            return ;
        }
        err = at_tok_nextstr(&line,&data);
        if(err != 0)
        {
            LOGE("at_tok_nextstr[data] error");
            onUssdResponse(mode);
            return ;
        }
        err = at_tok_nextint(&line,&dcs);
        if(err != 0)
        {
            LOGE("at_tok_nextstr[dcs] error");
            return ;
        }
        dcs &= 0x0c;

        response[0] = (char *)malloc(strlen(mode) + 1);
        memcpy(response[0],mode,strlen(mode) + 1);

        response[1] = (char *)malloc(strlen(data) + 1);

#ifdef CUSD_USE_UCS2_MODE
        if(true)//UCS2
#else
        if(dcs == 0x08)//UCS2
#endif
        {
            //unsigned char *tmp = (unsigned char *)malloc(sizeof(data)/2*3);
            size_t i = 0;
            unsigned char tmp[512];
            while(i < strlen(data))
            {   
                tmp[i/2] = (unsigned char)gsm_hex2_to_byte(&data[i]);
                i += 2;
            }   
            outlen = ucs2_to_utf8(tmp,i/4,(bytes_t)response[1]);
            response[1][outlen] = 0;
            //free(tmp);
        }
#if 0 // CUSD_USE_UCS2_MODE
        else if(dcs == 0x04)// 8 bit data
        {
            //memcpy(response[1],data,strlen(data) + 1);
            outlen = gsm_hex_to_bytes((bytes_t)data,strlen(data),(bytes_t)response[1]);
        }
        else//GSM 7 bit data, infact module gives us GSM 8 bit data
        {
            // memcpy(response[1],data,strlen(data) + 1);
            outlen = utf8_from_gsm8((cbytes_t)data,strlen(data),(bytes_t)response[1]);
        }
#endif
        LOGI("len:%d response[0]:%s",strlen(response[0]),response[0]);
        LOGI("len:%d[%d] response[1]:%s",strlen(response[1]),outlen,response[1]);

        RIL_onUnsolicitedResponse(RIL_UNSOL_ON_USSD, response, sizeof(response[0]) + sizeof(response[1]));

        free(response[0]);
        free(response[1]);
    } 
#endif
}

/* Called on command or reader thread */
static void onATReaderClosed()
{
    LOGI("AT channel closed\n");
    at_close();
    s_closed = 1;

    setRadioState (RADIO_STATE_UNAVAILABLE);
}

/* Called on command thread */
static void onATTimeout()
{
    LOGI("AT channel timeout; closing\n");
    at_close();

    s_closed = 1;

    /* FIXME cause a radio reset here */

    setRadioState (RADIO_STATE_UNAVAILABLE);
}

static void usage(char *s)
{
#ifdef RIL_SHLIB
    fprintf(stderr, "reference-ril requires: -p <tcp port> or -d /dev/tty_device\n");
#else
    fprintf(stderr, "usage: %s [-p <tcp port>] [-d /dev/tty_device]\n", s);
    exit(-1);
#endif
}

static void *
mainLoop(void *param)
{
    int fd;
    int ret;
    int get_at_port_fail_times = 100;

#if 0 //quectel //for debug-purpose, record logcat msg to file
//you can fetch logfiles to host-pc by adb tools using command "adb pull /data/ql_log/log_file_name"
    char logcat_cmd[100];
    time_t rawtime;
    struct tm *timeinfo;	
    time(&rawtime);
    timeinfo=localtime(&rawtime );
    system("mkdir /data/ql_log");
    sprintf(logcat_cmd, "logcat -v time -f /data/ql_log/%02d%02d_%02d%02d%02d_main.txt &",
		timeinfo->tm_mon+1, timeinfo->tm_mday, timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec);
    system(logcat_cmd);
    sprintf(logcat_cmd, "logcat -b radio -v time -f /data/ql_log/%02d%02d_%02d%02d%03d_radio.txt &",
		timeinfo->tm_mon+1, timeinfo->tm_mday, timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec);
    system(logcat_cmd);
#endif

    LOGD("Quectel RIL Version: " REFERENCE_RIL_VERSION);
    LOGI("mainLoop Start\r\n");
    AT_DUMP("== ", "entering mainLoop()", -1 );
    at_set_on_reader_closed(onATReaderClosed);
    at_set_on_timeout(onATTimeout);

    for (;;) {
#if 1 //quectel
        if (s_device_path == ql_ttyAT) {
            char atdevice[10];
__get_at_port_start:
            //1. can not get at port for long time, maybe quectel-module is power-off
            //2. when AT timeout occurs, need to restart quectel-module
            if (get_at_port_fail_times > 5) {
                // turn off quectel-module's power-supply here
                //sleep(2); //wait power-supply stable
                //turn on quectel-module's power-supply here
                //sleep(2); //wait power-supply stable
                //use pwr-key to turn on quectel-module here
                //sleep(5); //wait quectel-power boot up
                get_at_port_fail_times = 0;
            }
            if (ql_get_ttyAT(atdevice) != 0) {
                LOGE("cannot find ttyAT");
                get_at_port_fail_times++;
                sleep(3);
                goto __get_at_port_start;
            }
            sprintf(ql_ttyAT, "/dev/%s", atdevice);
            LOGD("quectel at port is %s", ql_ttyAT);
            get_at_port_fail_times = 0;
        }
#endif       

        fd = -1;
        while  (fd < 0) {
            if (s_port > 0) {
                fd = socket_loopback_client(s_port, SOCK_STREAM);
            } else if (s_device_socket) {
                if (!strcmp(s_device_path, "/dev/socket/qemud")) {
                    /* Before trying to connect to /dev/socket/qemud (which is
                     * now another "legacy" way of communicating with the
                     * emulator), we will try to connecto to gsm service via
                     * qemu pipe. */
                    fd = qemu_pipe_open("qemud:gsm");
                    if (fd < 0) {
                        /* Qemu-specific control socket */
                        fd = socket_local_client( "qemud",
                                                  ANDROID_SOCKET_NAMESPACE_RESERVED,
                                                  SOCK_STREAM );
                        if (fd >= 0 ) {
                            char  answer[2];

                            if ( write(fd, "gsm", 3) != 3 ||
                                 read(fd, answer, 2) != 2 ||
                                 memcmp(answer, "OK", 2) != 0)
                            {
                                close(fd);
                                fd = -1;
                            }
                       }
                    }
                }
                else
                    fd = socket_local_client( s_device_path,
                                            ANDROID_SOCKET_NAMESPACE_FILESYSTEM,
                                            SOCK_STREAM );
            } else if (s_device_path != NULL) {
                fd = open (s_device_path, O_RDWR);
#if 1 //quectel
                if ( fd >= 0 ) {
                    /* disable echo on serial ports */
                    struct termios  ios;
                    tcgetattr( fd, &ios );
                    cfmakeraw(&ios);
                    ios.c_lflag = 0;  /* disable ECHO, ICANON, etc... */
                    cfsetispeed(&ios, B115200);
                    cfsetospeed(&ios, B115200);
                    tcsetattr( fd, TCSANOW, &ios );
                    LOGD("open device %s correctly\n", s_device_path);
                    tcflush(fd, TCIOFLUSH);
                }
#else
                if ( fd >= 0 && !memcmp( s_device_path, "/dev/ttyS", 9 ) ) {
                    /* disable echo on serial ports */
                    struct termios  ios;
                    tcgetattr( fd, &ios );
                    ios.c_lflag = 0;  /* disable ECHO, ICANON, etc... */
                    tcsetattr( fd, TCSANOW, &ios );
                }
#endif
            }

            if (fd < 0) {
                LOGE("open device %s error for %s\n", s_device_path, strerror(errno));
                perror ("opening AT interface. retrying...");
#if 1 //quectel
                sleep(3);
#else
                sleep(10);
#endif
                /* never returns */
            }
        }

        s_closed = 0;
        ret = at_open(fd, onUnsolicited);

        if (ret < 0) {
            LOGE ("AT error %d on at_open\n", ret);
            return 0;
        }

        RIL_requestTimedCallback(initializeCallback, NULL, &TIMEVAL_0);

        // Give initializeCallback a chance to dispatched, since
        // we don't presently have a cancellation mechanism
        sleep(1);

        waitForClose();
#if 1 //quectel
        kill_PPP();
        get_at_port_fail_times = 100; //force to restart quectel-module
#endif
        LOGI("Re-opening after close");
    }
}

#ifdef RIL_SHLIB

pthread_t s_tid_mainloop;

const RIL_RadioFunctions *RIL_Init(const struct RIL_Env *env, int argc, char **argv)
{
    int ret;
    int fd = -1;
    int opt;
    pthread_attr_t attr;

    s_rilenv = env;

    while ( -1 != (opt = getopt(argc, argv, "p:d:s:"))) {
        switch (opt) {
            case 'p':
                s_port = atoi(optarg);
                if (s_port == 0) {
                    usage(argv[0]);
                    return NULL;
                }
                LOGI("Opening loopback port %d\n", s_port);
            break;

            case 'd':
                s_device_path = optarg;
                LOGI("Opening tty device %s\n", s_device_path);
            break;

            case 's':
                s_device_path   = optarg;
                s_device_socket = 1;
                LOGI("Opening socket %s\n", s_device_path);
            break;

            default:
                usage(argv[0]);
                return NULL;
        }
    }

    if (s_port < 0 && s_device_path == NULL) {
#if 1 //quectel //get ttyAT dymanic
    s_device_path = ql_ttyAT;
#else
        usage(argv[0]);
        return NULL;
#endif
    }

    pthread_attr_init (&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    ret = pthread_create(&s_tid_mainloop, &attr, mainLoop, NULL);

    return &s_callbacks;
}
#else /* RIL_SHLIB */
int main (int argc, char **argv)
{
    int ret;
    int fd = -1;
    int opt;

    while ( -1 != (opt = getopt(argc, argv, "p:d:"))) {
        switch (opt) {
            case 'p':
                s_port = atoi(optarg);
                if (s_port == 0) {
                    usage(argv[0]);
                }
                LOGI("Opening loopback port %d\n", s_port);
            break;

            case 'd':
                s_device_path = optarg;
                LOGI("Opening tty device %s\n", s_device_path);
            break;

            case 's':
                s_device_path   = optarg;
                s_device_socket = 1;
                LOGI("Opening socket %s\n", s_device_path);
            break;

            default:
                usage(argv[0]);
        }
    }

    if (s_port < 0 && s_device_path == NULL) {
        usage(argv[0]);
    }

    RIL_register(&s_callbacks);

    mainLoop(NULL);

    return 0;
}

#endif /* RIL_SHLIB */

